import logging
import time

from boto import exception as boto_exception
from botocore import exceptions as botocore_exceptions
from tornado import concurrent
from tornado import gen
from tornado import testing

from kingpin.actors.aws import api_call_queue

log = logging.getLogger(__name__)


class TestApiCallQueue(testing.AsyncTestCase):
    boto2_exception = boto_exception.BotoServerError(
        '400', 'Bad request',
        {'Error': {'Code': 'Bad request'}})
    boto2_throttle_exception_1 = boto_exception.BotoServerError(
        '429', 'Rate limit',
        {'Error': {'Code': 'Throttling'}})
    boto2_throttle_exception_2 = boto_exception.BotoServerError(
        '429', 'Rate limit',
        {'Error': {'Code': 'Rate exceeded'}})
    boto2_throttle_exception_3 = boto_exception.BotoServerError(
        '429', 'Rate limit',
        {'Error': {'Code': 'reached max retries'}})

    boto3_exception = botocore_exceptions.ClientError(
        {'Error': {'Code': 'Bad request'}}, 'Test')
    boto3_throttle_exception = botocore_exceptions.ClientError(
        {'Error': {'Code': 'Throttling'}}, 'Test')

    def setUp(self):
        super(TestApiCallQueue, self).setUp()
        self.api_call_queue = api_call_queue.ApiCallQueue()
        self.api_call_queue.delay_min = 0.05
        self.api_call_queue.delay_max = 0.2

        self.executor = concurrent.futures.ThreadPoolExecutor(10)

    @testing.gen_test
    def test_plain_call(self):
        """Test that a single api call through the queue works."""
        result = yield self.api_call_queue.call(self._mock_api_function_sync)
        self.assertEqual(result, 'OK')

    @testing.gen_test
    def test_concurrent_calls_with_delay(self):
        """
        Test concurrent calls with some latency run serially
        and return independent results.

        The api_call_queue runs calls synchronously and serially.
        """
        api_call_queue_calls = [
            self.api_call_queue.call(
                self._mock_api_function_sync, result=1, delay=0.05),
            self.api_call_queue.call(
                self._mock_api_function_sync, result=2, delay=0.05),
            self.api_call_queue.call(
                self._mock_api_function_sync, result=3, delay=0.05),
            self.api_call_queue.call(
                self._mock_api_function_sync, result=4, delay=0.05),
            self.api_call_queue.call(
                self._mock_api_function_sync, result=5, delay=0.05),
        ]

        start = time.time()
        results = yield gen.multi(api_call_queue_calls)
        stop = time.time()
        run_time = stop - start

        self.assertTrue(0.25 <= run_time < 0.35)
        self.assertEqual(results, [1, 2, 3, 4, 5])

    @testing.gen_test
    def test_api_call_queue_future_is_nonblocking(self):
        """
        Test that the api call queue future is nonblocking for other futures.
        It needs to execute in a different thread because of the synchronous
        nature of AWS APIs.
        """
        futures = [
            self.api_call_queue.call(
                self._mock_api_function_sync, result=1, delay=0.05),
            self._mock_api_function_async(result=2, delay=0.05),
            self._mock_api_function_async(result=3, delay=0.05),
            self._mock_api_function_async(result=4, delay=0.05),
            self._mock_api_function_async(result=5, delay=0.05),
        ]

        start = time.time()
        results = yield gen.multi(futures)
        stop = time.time()
        run_time = stop - start

        self.assertTrue(0.05 <= run_time < 0.15)
        self.assertEqual(results, [1, 2, 3, 4, 5])

    @testing.gen_test
    def test_api_call_queue_raises_exceptions(self):
        """
        Test that the api call queue raises exceptions and proceeds to execute
        other queued api calls.
        """

        @gen.coroutine
        def _call_without_exception():
            result = yield self.api_call_queue.call(
                self._mock_api_function_sync, delay=0.05)
            self.assertEqual(result, 'OK')
            raise gen.Return('no exception')

        @gen.coroutine
        def _call_with_exception():
            err = ValueError('test exception')
            try:
                yield self.api_call_queue.call(
                    self._mock_api_function_sync,
                    exception=err,
                    delay=0.05)
            except Exception as e:
                self.assertEqual(err, e)
            raise gen.Return('exception')

        @gen.coroutine
        def _call_with_exception_after_boto2_rate_limit():
            """
            First rate limit, then raise an exception.
            This should take:
                call delay * 2 + min rate limiting delay * 1
            """
            try:
                yield self.api_call_queue.call(
                    self._mock_api_function_sync,
                    exception=[
                        self.boto2_throttle_exception_1,
                        self.boto2_exception],
                    delay=0.05)
            except Exception as e:
                self.assertEqual(self.boto2_exception, e)

            raise gen.Return('exception')

        @gen.coroutine
        def _call_with_exception_after_boto3_rate_limit():
            """
            First rate limit, then raise an exception.
            This should take:
                call delay * 2 + min rate limiting delay * 1
            """
            try:
                yield self.api_call_queue.call(
                    self._mock_api_function_sync,
                    exception=[
                        self.boto3_throttle_exception,
                        self.boto3_exception],
                    delay=0.05)
            except Exception as e:
                self.assertEqual(self.boto3_exception, e)

            raise gen.Return('exception')

        call_wrappers = [
            # Should take 0.05s.
            _call_without_exception(),
            # Should take 0.05s.
            _call_with_exception(),
            # Should take 0.05s.
            _call_without_exception(),
            # Should take 0.05s + 0.05s + 0.05s.
            _call_with_exception_after_boto2_rate_limit(),
            # Should take 0.05s + 0.05s + 0.05s.
            _call_with_exception_after_boto3_rate_limit(),
        ]

        start = time.time()
        results = yield gen.multi_future(call_wrappers)
        stop = time.time()
        run_time = stop - start

        self.assertTrue(0.45 <= run_time < 0.55)
        self.assertEqual(
            results,
            ['no exception', 'exception', 'no exception',
             'exception', 'exception'])

    @testing.gen_test
    def test_rate_limiting_boto2(self):
        """
        Test that rate limiting with boto2 works.
        """
        # Each one of these will raise a throttling exception, then succeed.
        # Between calls, each should delay for 1 cycle of `delay_min`.
        # Delay min is 0.05s, so total runtime should be ~0.15s.
        api_call_queue_calls = [
            self.api_call_queue.call(
                self._mock_api_function_sync,
                result=1,
                exception=[self.boto2_throttle_exception_1]),
            self.api_call_queue.call(
                self._mock_api_function_sync,
                result=2,
                exception=[self.boto2_throttle_exception_2]),
            self.api_call_queue.call(
                self._mock_api_function_sync,
                result=3,
                exception=[self.boto2_throttle_exception_3]),
        ]

        start = time.time()
        results = yield gen.multi(api_call_queue_calls)
        stop = time.time()
        run_time = stop - start

        self.assertTrue(0.15 <= run_time < 0.25)
        self.assertEqual(results, [1, 2, 3])

    @testing.gen_test
    def test_rate_limiting_boto3(self):
        """
        Test that rate limiting with boto3 works.
        """
        # Each one of these will raise a throttling exception, then succeed.
        # Between calls, each should delay for 1 cycle of `delay_min`.
        # Delay min is 0.05s, so total runtime should be ~0.15s.
        api_call_queue_calls = [
            self.api_call_queue.call(
                self._mock_api_function_sync,
                result=1,
                exception=[self.boto3_throttle_exception]),
            self.api_call_queue.call(
                self._mock_api_function_sync,
                result=2,
                exception=[self.boto3_throttle_exception]),
            self.api_call_queue.call(
                self._mock_api_function_sync,
                result=3,
                exception=[self.boto3_throttle_exception]),
        ]

        start = time.time()
        results = yield gen.multi(api_call_queue_calls)
        stop = time.time()
        run_time = stop - start

        self.assertTrue(0.15 <= run_time < 0.25)
        self.assertEqual(results, [1, 2, 3])

    @testing.gen_test
    def test_rate_limit_stepping(self):
        """
        Test that rate limiting steps delay up and down.
        """

        # The delay will step from delay_min to delay_max by doubling.
        # 0s -> 0.05s -> 0.10s -> 0.20s.
        # When a call succeeds, the delay goes back down a step.

        def _throttle_twice(result):
            # This will raise two throttling exception, then succeed.
            # Because of that, this should end with the delay going up
            # one step.
            return self.api_call_queue.call(
                self._mock_api_function_sync,
                result=result,
                exception=[
                    self.boto2_throttle_exception_1,
                    self.boto3_throttle_exception])

        api_call_queue_calls = [
            _throttle_twice(result=1),
        ]

        start = time.time()
        results = yield gen.multi(api_call_queue_calls)
        stop = time.time()
        run_time = stop - start

        # Delay should be up one step total: delay_min.

        self.assertEqual(
            self.api_call_queue.delay,
            self.api_call_queue.delay_min)
        self.assertTrue(0.15 <= run_time < 0.25)
        self.assertEqual(results, [1])

        # Do it again, to go up one more step.

        api_call_queue_calls = [
            _throttle_twice(result=2),
        ]

        start = time.time()
        results = yield gen.multi(api_call_queue_calls)
        stop = time.time()
        run_time = stop - start

        # Delay should be up two steps total: delay_min * 2.

        self.assertEqual(
            self.api_call_queue.delay,
            self.api_call_queue.delay_min * 2)
        self.assertTrue(0.35 <= run_time < 0.45)
        self.assertEqual(results, [2])

        # Do it again, to go up one more step.

        api_call_queue_calls = [
            _throttle_twice(result=3),
        ]

        start = time.time()
        results = yield gen.multi(api_call_queue_calls)
        stop = time.time()
        run_time = stop - start

        # Delay should be up to delay_min * 2 again,
        # because it always goes down once it succeeds,
        # but total runtime should be ~0.5s:
        # (0delay_min * 2 + delay_max + delay_max).

        self.assertEqual(
            self.api_call_queue.delay,
            self.api_call_queue.delay_min * 2)
        self.assertTrue(0.5 <= run_time < 0.6)
        self.assertEqual(results, [3])

        # Now do it with successes to step back down.

        api_call_queue_calls = [
            self.api_call_queue.call(
                self._mock_api_function_sync, result=4),
            self.api_call_queue.call(
                self._mock_api_function_sync, result=5),
            self.api_call_queue.call(
                self._mock_api_function_sync, result=6),
        ]

        start = time.time()
        results = yield gen.multi(api_call_queue_calls)
        stop = time.time()
        run_time = stop - start

        # Delay should be up to 0 again.
        # Total runtime should be ~0.15s:
        # (delay_min * 2 + delay_min).

        self.assertEqual(self.api_call_queue.delay, 0)
        self.assertTrue(0.15 <= run_time < 0.25)
        self.assertEqual(results, [4, 5, 6])

    def _mock_api_function_sync(self, result='OK',
                                exception=None,
                                delay=None):
        """
        Mock a synchronous call to AWS.

        Args:
            result: Value to return.
            exception:
                If this is an exception:

                Raise instead of returning.

                If this is a list:

                Raise and pop off the first exception
                instead of returning.
                On subsequent calls, the next exception will be raised.
                If the list is exhausted, it will not raise an exception.
            delay: If set, delay before returning a result or exception.

        Returns:
            This will return `result`.
            If `result` is not set, this will return the string 'OK'.
        """
        if delay is not None:
            time.sleep(delay)
        if exception is not None:
            if isinstance(exception, Exception):
                raise exception
            elif len(exception) > 0:
                raise exception.pop(0)

        return result

    @concurrent.run_on_executor
    def _mock_api_function_async(self, *args, **kwargs):
        """
        Wrapper around _mock_api_function_sync that runs it on another thread.
        """
        return self._mock_api_function_sync(*args, **kwargs)



from boto.exception import BotoServerError
from tornado import testing

from kingpin.actors.aws import settings


class TestSettings(testing.AsyncTestCase):

    def test_is_retriable(self):

        class TransientError(BotoServerError):
            """Unit-test exception"""

        exc = TransientError('500', 'Throttling everything')
        exc.error_code = 'Throttling'
        self.assertTrue(settings.is_retriable_exception(exc))

        self.assertFalse(settings.is_retriable_exception(Exception()))



import datetime
import logging
import json

from botocore.exceptions import ClientError
from tornado import testing
import mock

from kingpin.actors.aws import base
from kingpin.actors.aws import settings
from kingpin.actors.aws import cloudformation
from kingpin.actors.test.helper import tornado_value
import importlib

log = logging.getLogger(__name__)


def create_fake_stack(name, status):
    fake_stack = {
        'StackId': 'arn:aws:cloudformation:us-east-1:xxxx:stack/%s/x' % name,
        'LastUpdatedTime': datetime.datetime.now(),
        'TemplateDescription': 'Fake Template %s' % name,
        'CreationTime': datetime.datetime.now(),
        'StackName': name,
        'StackStatus': status,
        'StackStatusReason': 'Fake Reason',
        'EnableTerminationProtection': False,
        'Parameters': [
            {'ParameterKey': 'key1', 'ParameterValue': 'value1'}
        ],
    }
    return fake_stack


def create_fake_stack_event(name, resource, status, reason=None):
    fake_event = {
        'EventId': '264322b0-2426-11e6-aaa1-500c28b32ed2',
        'LogicalResourceId': resource,
        'PhysicalResourceId': 'arn:aws:cf:us-east-1:x:stack/%s/abc' % name,
        'ResourceStatus': status,
        'ResourceType': 'AWS::CloudFormation::Stack',
        'StackId': 'arn:aws:cf:us-east-1:xxx:stack/%s/xyz' % name,
        'StackName': name,
        'Timestamp': datetime.datetime.now(),
    }

    if reason:
        fake_event['ResourceStatusReason'] = reason

    return fake_event


class TestCloudFormationBaseActor(testing.AsyncTestCase):

    def setUp(self):
        super(TestCloudFormationBaseActor, self).setUp()
        settings.AWS_ACCESS_KEY_ID = 'unit-test'
        settings.AWS_SECRET_ACCESS_KEY = 'unit-test'
        settings.RETRYING_SETTINGS = {'stop_max_attempt_number': 1}
        importlib.reload(cloudformation)

        self.actor = cloudformation.CloudFormationBaseActor(
            'unittest', {'region': 'us-east-1'})
        self.actor.cf3_conn = mock.MagicMock(name='cf3_conn')

        # Need to recreate the api call queues between tests
        # because nose creates a new ioloop per test run.
        base.NAMED_API_CALL_QUEUES = {}

    def test_discover_noecho_params(self):
        file = 'examples/test/aws.cloudformation/cf.integration.json'
        (body, url) = self.actor._get_template_body(file)
        ret = self.actor._discover_noecho_params(body)
        self.assertEqual(ret, ['BucketPassword'])

    def test_get_template_body(self):
        file = 'examples/test/aws.cloudformation/cf.unittest.json'
        url = 'http://foobar.json'

        # Should work...
        ret = self.actor._get_template_body(file)
        expected = ('{"blank": "json"}', None)
        self.assertEqual(ret, expected)

        # Should return None
        ret = self.actor._get_template_body(None)
        expected = (None, None)
        self.assertEqual(ret, expected)

        # Should return None
        ret = self.actor._get_template_body(url)
        expected = (None, 'http://foobar.json')
        self.assertEqual(ret, expected)

        # Should raise exception
        with self.assertRaises(cloudformation.InvalidTemplate):
            self.actor._get_template_body('missing')

    @testing.gen_test
    def test_validate_template_body(self):
        yield self.actor._validate_template(body='test body')
        self.actor.cf3_conn.validate_template.assert_called_with(
            TemplateBody='test body')

    @testing.gen_test
    def test_validate_template_url(self):
        yield self.actor._validate_template(url='http://foobar.json')
        self.actor.cf3_conn.validate_template.assert_called_with(
            TemplateURL='http://foobar.json')

    @testing.gen_test
    def test_validate_template_raises_boto_error(self):
        fake_exc = {
            'ResponseMetadata': {
                'HTTPStatusCode': 400,
                'RequestId': 'dfc1a12c-22c1-11e6-80b1-8fd4cf167f54'
            },
            'Error': {
                'Message': 'Template format error: JSON not well-formed',
                'Code': 'ValidationError',
                'Type': 'Sender'
            }
        }

        self.actor.cf3_conn.validate_template.side_effect = ClientError(
            fake_exc, 'FakeOperation')
        with self.assertRaises(cloudformation.InvalidTemplate):
            yield self.actor._validate_template(url='junk')

        with self.assertRaises(cloudformation.InvalidTemplate):
            yield self.actor._validate_template(body='junk')

    def test_create_parameters(self):
        params = {
            'Key1': 'Value1',
            'Key2': 'Value2',
        }

        expected = [
            {'ParameterKey': 'Key1',
             'ParameterValue': 'Value1'},
            {'ParameterKey': 'Key2',
             'ParameterValue': 'Value2'}
        ]

        actor = cloudformation.CloudFormationBaseActor(
            'unittest', {'region': 'us-east-1'})

        ret = actor._create_parameters(params)
        self.assertEqual(ret, expected)

    @testing.gen_test
    def test_get_stack(self):
        self.actor.cf3_conn.describe_stacks.return_value = {
            'Stacks': [create_fake_stack('s1', 'UPDATE_COMPLETE')]}

        ret = yield self.actor._get_stack('s1')
        self.assertEqual(ret['StackName'], 's1')

    @testing.gen_test
    def test_get_stack_not_found(self):
        fake_exc = {
            'ResponseMetadata': {
                'HTTPStatusCode': 400,
                'RequestId': 'dfc1a12c-22c1-11e6-80b1-8fd4cf167f54'
            },
            'Error': {
                'Message': 'Stack with id s1 does not exist',
                'Code': 'ExecutionFailure',
                'Type': 'Sender'
            }
        }
        self.actor.cf3_conn.describe_stacks.side_effect = ClientError(
            fake_exc, 'Failure')

        ret = yield self.actor._get_stack('s1')
        self.assertEqual(ret, None)

    @testing.gen_test
    def test_get_stack_exc(self):
        fake_exc = {
            'ResponseMetadata': {
                'HTTPStatusCode': 400,
                'RequestId': 'dfc1a12c-22c1-11e6-80b1-8fd4cf167f54'
            },
            'Error': {
                'Message': 'Some other error',
                'Code': 'ExecutionFailure',
                'Type': 'Sender'
            }
        }
        self.actor.cf3_conn.describe_stacks.side_effect = ClientError(
            fake_exc, 'Failure')

        with self.assertRaises(cloudformation.CloudFormationError):
            yield self.actor._get_stack('s1')

    @testing.gen_test
    def test_get_stack_template(self):
        fake_stack_template = {
            'ResponseMetadata': {
                'HTTPStatusCode': 200,
                'RequestId': '6dcafb4c-2768-11e6-b748-295d1284a76f'
            },
            'TemplateBody': {
                'Fake': 'Stack'
            }
        }
        self.actor.cf3_conn.get_template.return_value = fake_stack_template
        ret = yield self.actor._get_stack_template('test')
        self.actor.cf3_conn.get_template.assert_has_calls(
            [mock.call(StackName='test', TemplateStage='Original')])
        self.assertEqual(ret, {'Fake': 'Stack'})

    @testing.gen_test
    def test_get_stack_template_exc(self):
        fake_exc = {
            'ResponseMetadata': {
                'HTTPStatusCode': 400,
                'RequestId': 'dfc1a12c-22c1-11e6-80b1-8fd4cf167f54'
            },
            'Error': {
                'Message': 'Some other error',
                'Code': 'ExecutionFailure',
                'Type': 'Sender'
            }
        }
        self.actor.cf3_conn.get_template.side_effect = ClientError(
            fake_exc, 'Failure')
        with self.assertRaises(cloudformation.CloudFormationError):
            yield self.actor._get_stack_template('test')

    @testing.gen_test
    def test_wait_until_state_complete(self):
        create_in_progress = create_fake_stack('test', 'CREATE_IN_PROGRESS')
        create_complete = create_fake_stack('test', 'CREATE_COMPLETE')

        # Make _get_stack() yield back 2 in-progress states, then yield a
        # successfull execution.
        self.actor._get_stack = mock.MagicMock(name='FakeStack')
        self.actor._get_stack.side_effect = [
            tornado_value(create_in_progress),
            tornado_value(create_in_progress),
            tornado_value(create_complete)
        ]
        yield self.actor._wait_until_state(
            'test', cloudformation.COMPLETE, sleep=0.01)
        self.actor._get_stack.assert_has_calls(
            [mock.call('test'), mock.call('test'), mock.call('test')])

    @testing.gen_test
    def test_wait_until_state_stack_failed(self):
        create_in_progress = create_fake_stack('test', 'CREATE_IN_PROGRESS')
        create_complete = create_fake_stack('test', 'CREATE_COMPLETE')

        # Make sure a cloudformationerror is raised if we ask for a deleted
        # state rather than a created one.
        self.actor._get_stack = mock.MagicMock(name='FakeStack')
        self.actor._get_stack.side_effect = [
            tornado_value(create_in_progress),
            tornado_value(create_in_progress),
            tornado_value(create_complete)
        ]
        with self.assertRaises(cloudformation.StackFailed):
            yield self.actor._wait_until_state(
                'test', cloudformation.DELETED, sleep=0.1)

    @testing.gen_test
    def test_wait_until_state_stack_not_found(self):
        # Lastly, test that if wait_until_state returns no actor, we bail
        # appropriately.
        self.actor._get_stack = mock.MagicMock(name='FakeStack')
        self.actor._get_stack.return_value = tornado_value(None)
        with self.assertRaises(cloudformation.StackNotFound):
            yield self.actor._wait_until_state(
                'test', cloudformation.COMPLETE, sleep=0.1)

    @testing.gen_test
    def test_get_stack_events(self):
        fake_events = {
            'StackEvents': [
                create_fake_stack_event('test', 'test', 'DELETE_COMPLETE'),
                create_fake_stack_event('test', 's3', 'DELETE_COMPLETE'),
                create_fake_stack_event('test', 's3', 'DELETE_IN_PROGRESS'),
                create_fake_stack_event('test', 's3', 'CREATE_FAILED', 'bad'),
                create_fake_stack_event('test', 's3', 'CREATE_IN_PROGRESS'),
                create_fake_stack_event('test', 'test', 'CREATE_IN_PROGRESS')
            ]
        }
        expected = [
            'AWS::CloudFormation::Stack test (CREATE_IN_PROGRESS): ',
            'AWS::CloudFormation::Stack s3 (CREATE_IN_PROGRESS): ',
            'AWS::CloudFormation::Stack s3 (CREATE_FAILED): bad',
            'AWS::CloudFormation::Stack s3 (DELETE_IN_PROGRESS): ',
            'AWS::CloudFormation::Stack s3 (DELETE_COMPLETE): ',
            'AWS::CloudFormation::Stack test (DELETE_COMPLETE): '
        ]
        self.actor.cf3_conn.describe_stack_events.return_value = fake_events
        ret = yield self.actor._get_stack_events('test')

        self.assertEqual(ret, expected)

    @testing.gen_test
    def test_get_stack_events_exc(self):
        fake_exc = {
            'ResponseMetadata': {
                'HTTPStatusCode': 400,
                'RequestId': 'dfc1a12c-22c1-11e6-80b1-8fd4cf167f54'
            },
            'Error': {
                'Message': 'Some other error',
                'Code': 'ExecutionFailure',
                'Type': 'Sender'
            }
        }
        self.actor.cf3_conn.describe_stack_events.side_effect = ClientError(
            fake_exc, 'Failure')
        ret = yield self.actor._get_stack_events('test')
        self.assertEqual(ret, [])

    @testing.gen_test
    def test_delete_stack(self):
        self.actor.cf3_conn.delete_stack.return_value = {
            'ResponseMetadata': {'RequestId': 'req-id-1'}
        }
        self.actor._wait_until_state = mock.MagicMock(name='_wait_until_state')
        exc = cloudformation.StackNotFound()
        self.actor._wait_until_state.side_effect = exc

        yield self.actor._delete_stack(stack='stack')

        self.assertTrue(self.actor.cf3_conn.delete_stack.called)
        self.assertTrue(self.actor._wait_until_state.called)

    @testing.gen_test
    def test_delete_stack_raises_boto_error(self):
        self.actor.cf3_conn.delete_stack = mock.MagicMock(name='delete_stack')

        fake_exc = {
            'ResponseMetadata': {
                'HTTPStatusCode': 400,
                'RequestId': 'dfc1a12c-22c1-11e6-80b1-8fd4cf167f54'
            },
            'Error': {
                'Message': 'Something failed',
                'Code': 'ExecutionFailure',
                'Type': 'Sender'
            }
        }

        self.actor.cf3_conn.delete_stack.side_effect = ClientError(
            fake_exc, 'Error')
        with self.assertRaises(cloudformation.CloudFormationError):
            yield self.actor._delete_stack(stack='stack')


class TestCreate(testing.AsyncTestCase):

    def setUp(self):
        super(TestCreate, self).setUp()
        settings.AWS_ACCESS_KEY_ID = 'unit-test'
        settings.AWS_SECRET_ACCESS_KEY = 'unit-test'
        settings.RETRYING_SETTINGS = {'stop_max_attempt_number': 1}
        importlib.reload(cloudformation)
        # Need to recreate the api call queues between tests
        # because nose creates a new ioloop per test run.
        base.NAMED_API_CALL_QUEUES = {}

    @testing.gen_test
    def test_create_stack_file(self):
        actor = cloudformation.Create(
            'Unit Test Action',
            {'name': 'unit-test-cf',
             'region': 'us-west-2',
             'template':
                 'examples/test/aws.cloudformation/cf.integration.json'})
        actor._wait_until_state = mock.MagicMock(name='_wait_until_state')
        actor._wait_until_state.side_effect = [tornado_value(None)]
        actor.cf3_conn.create_stack = mock.MagicMock(name='create_stack_mock')
        actor.cf3_conn.create_stack.return_value = {'StackId': 'arn:123'}
        ret = yield actor._create_stack(stack='test')
        self.assertEqual(ret, 'arn:123')

    @testing.gen_test
    def test_create_stack_file_with_role(self):
        stack = 'examples/test/aws.cloudformation/cf.integration.json'
        actor = cloudformation.Create(
            'Unit Test Action',
            {'name': 'unit-test-cf',
             'region': 'us-west-2',
             'role_arn': 'test_role_arn',
             'template': stack})
        actor._wait_until_state = mock.MagicMock(name='_wait_until_state')
        actor._wait_until_state.side_effect = [tornado_value(None)]
        actor.cf3_conn.create_stack = mock.MagicMock(name='create_stack_mock')
        actor.cf3_conn.create_stack.return_value = {'StackId': 'arn:123'}
        ret = yield actor._create_stack(stack='test')
        self.assertEqual(ret, 'arn:123')
        actor.cf3_conn.create_stack.assert_called_with(
            TemplateBody=mock.ANY,
            EnableTerminationProtection=False,
            Parameters=[],
            RoleARN='test_role_arn',
            TimeoutInMinutes=60,
            Capabilities=[],
            StackName='test',
            OnFailure='DELETE')

    @testing.gen_test
    def test_create_stack_file_with_termination_protection_true(self):
        stack = 'examples/test/aws.cloudformation/cf.integration.json'
        actor = cloudformation.Create(
            'Unit Test Action',
            {'name': 'unit-test-cf',
             'region': 'us-west-2',
             'role_arn': 'test_role_arn',
             'template': stack})
        actor._options['enable_termination_protection'] = True
        actor._wait_until_state = mock.MagicMock(name='_wait_until_state')
        actor._wait_until_state.side_effect = [tornado_value(None)]
        actor.cf3_conn.create_stack = mock.MagicMock(name='create_stack_mock')
        actor.cf3_conn.create_stack.return_value = {'StackId': 'arn:123'}
        ret = yield actor._create_stack(stack='test')
        self.assertEqual(ret, 'arn:123')
        actor.cf3_conn.create_stack.assert_called_with(
            TemplateBody=mock.ANY,
            EnableTerminationProtection=True,
            Parameters=[],
            RoleARN='test_role_arn',
            TimeoutInMinutes=60,
            Capabilities=[],
            StackName='test',
            OnFailure='DELETE')

    @testing.gen_test
    def test_create_stack_url(self):
        actor = cloudformation.Create(
            'Unit Test Action',
            {'name': 'unit-test-cf',
             'region': 'us-west-2',
             'template': 'https://www.test.com'})
        actor._wait_until_state = mock.MagicMock(name='_wait_until_state')
        actor._wait_until_state.side_effect = [tornado_value(None)]
        actor.cf3_conn.create_stack = mock.MagicMock(name='create_stack_mock')
        actor.cf3_conn.create_stack.return_value = {'StackId': 'arn:123'}
        ret = yield actor._create_stack(stack='unit-test-cf')
        self.assertEqual(ret, 'arn:123')

    @testing.gen_test
    def test_create_stack_raises_boto_error(self):
        actor = cloudformation.Create(
            'Unit Test Action',
            {'name': 'unit-test-cf',
             'region': 'us-west-2',
             'template':
                 'examples/test/aws.cloudformation/cf.integration.json'})
        actor.cf3_conn.create_stack = mock.MagicMock(name='create_stack_mock')

        fake_exc = {
            'ResponseMetadata': {
                'HTTPStatusCode': 400,
                'RequestId': 'dfc1a12c-22c1-11e6-80b1-8fd4cf167f54'
            },
            'Error': {
                'Message': 'Something failed',
                'Code': 'ExecutionFailure',
                'Type': 'Sender'
            }
        }

        actor.cf3_conn.create_stack.side_effect = ClientError(
            fake_exc, 'Failure')
        with self.assertRaises(cloudformation.CloudFormationError):
            yield actor._create_stack(stack='test')

    @testing.gen_test
    def test_create_stack_wait_until_raises_boto_error(self):
        actor = cloudformation.Create(
            'Unit Test Action',
            {'name': 'unit-test-cf',
             'region': 'us-west-2',
             'template':
                 'examples/test/aws.cloudformation/cf.integration.json'})
        actor.cf3_conn.create_stack = mock.MagicMock(name='create_stack_mock')
        actor.cf3_conn.create_stack.return_value = {'StackId': 'arn:123'}

        actor._wait_until_state = mock.MagicMock(name='_wait_until_state')
        actor._wait_until_state.side_effect = cloudformation.StackFailed()

        actor._get_stack_events = mock.MagicMock(name='_get_stack_events')
        actor._get_stack_events.return_value = tornado_value(
            ['Log Message'])

        with self.assertRaises(cloudformation.StackFailed):
            yield actor._create_stack(stack='test')

    @testing.gen_test
    def test_execute(self):
        actor = cloudformation.Create(
            'Unit Test Action',
            {'name': 'unit-test-cf',
             'region': 'us-west-2',
             'template':
                 'examples/test/aws.cloudformation/cf.integration.json'})

        actor._validate_template = mock.MagicMock(name='_validate_template')
        actor._validate_template.return_value = tornado_value(True)

        actor._get_stack = mock.MagicMock(name='_get_stack')
        actor._get_stack.return_value = tornado_value(None)

        actor._create_stack = mock.MagicMock(name='_create_stack')
        actor._create_stack.return_value = tornado_value(None)

        actor._wait_until_state = mock.MagicMock(name='_wait_until_state')
        actor._wait_until_state.return_value = tornado_value(None)
        yield actor._execute()

    @testing.gen_test
    def test_execute_exists(self):
        actor = cloudformation.Create(
            'Unit Test Action',
            {'name': 'unit-test-cf',
             'region': 'us-west-2',
             'template':
                 'examples/test/aws.cloudformation/cf.integration.json'})

        actor._validate_template = mock.MagicMock(name='_validate_template')
        actor._validate_template.return_value = tornado_value(True)

        actor._get_stack = mock.MagicMock(name='_get_stack')
        actor._get_stack.return_value = tornado_value(True)

        with self.assertRaises(cloudformation.StackAlreadyExists):
            yield actor._execute()

    @testing.gen_test
    def test_execute_dry(self):
        actor = cloudformation.Create(
            'Unit Test Action',
            {'name': 'unit-test-cf',
             'region': 'us-west-2',
             'template':
                 'examples/test/aws.cloudformation/cf.integration.json'},
            dry=True)

        actor._validate_template = mock.MagicMock(name='_validate_template')
        actor._validate_template.return_value = tornado_value(True)

        actor._get_stack = mock.MagicMock(name='_get_stack')
        actor._get_stack.return_value = tornado_value(None)

        yield actor._execute()


class TestDelete(testing.AsyncTestCase):

    def setUp(self):
        super(TestDelete, self).setUp()
        settings.AWS_ACCESS_KEY_ID = 'unit-test'
        settings.AWS_SECRET_ACCESS_KEY = 'unit-test'
        settings.RETRYING_SETTINGS = {'stop_max_attempt_number': 1}
        importlib.reload(cloudformation)
        # Need to recreate the api call queues between tests
        # because nose creates a new ioloop per test run.
        base.NAMED_API_CALL_QUEUES = {}

    @testing.gen_test
    def test_execute(self):
        actor = cloudformation.Delete(
            'Unit Test Action',
            {'name': 'unit-test-cf',
             'region': 'us-west-2'})
        actor._get_stack = mock.MagicMock(name='_get_stack')
        actor._get_stack.return_value = tornado_value(True)
        actor._delete_stack = mock.MagicMock(name='_delete_stack')
        actor._delete_stack.return_value = tornado_value(None)
        actor._wait_until_state = mock.MagicMock(name='_wait_until_state')
        actor._wait_until_state.side_effect = cloudformation.StackNotFound()
        yield actor._execute()

    @testing.gen_test
    def test_execute_dry(self):
        actor = cloudformation.Delete(
            'Unit Test Action',
            {'name': 'unit-test-cf',
             'region': 'us-west-2'}, dry=True)
        actor._get_stack = mock.MagicMock(name='_get_stack')
        actor._get_stack.return_value = tornado_value(True)
        yield actor._execute()

    @testing.gen_test
    def test_execute_not_exists(self):
        actor = cloudformation.Delete(
            'Unit Test Action',
            {'name': 'unit-test-cf',
             'region': 'us-west-2'})
        actor._get_stack = mock.MagicMock(name='_get_stack')
        actor._get_stack.return_value = tornado_value(None)
        with self.assertRaises(cloudformation.StackNotFound):
            yield actor._execute()


class TestStack(testing.AsyncTestCase):

    def setUp(self):
        super(TestStack, self).setUp()
        settings.AWS_ACCESS_KEY_ID = 'unit-test'
        settings.AWS_SECRET_ACCESS_KEY = 'unit-test'
        settings.RETRYING_SETTINGS = {'stop_max_attempt_number': 1}
        importlib.reload(cloudformation)
        # Need to recreate the api call queues between tests
        # because nose creates a new ioloop per test run.
        base.NAMED_API_CALL_QUEUES = {}

        self.actor = cloudformation.Stack(
            options={
                'name': 'unit-test-cf',
                'state': 'present',
                'region': 'us-west-2',
                'template':
                    'examples/test/aws.cloudformation/cf.unittest.json',
                'parameters': {
                    'key1': 'value1'
                }
            })
        self.actor.cf3_conn = mock.MagicMock(name='cf3_conn')

    def test_diff_params_safely(self):
        self.actor = cloudformation.Stack(
            options={
                'name': 'unit-test-cf',
                'state': 'present',
                'region': 'us-west-2',
                'template':
                    'examples/test/aws.cloudformation/cf.integration.json',
                'parameters': {
                    'BucketName': 'name',
                    'BucketPassword': 'test_password',
                    'Metadata': '1.0'
                }
            })

        # Pretend that the parameters are the same
        # (BucketName, DefaultParam, and Metadata), and the BucketPassword
        # came back with stars.
        # We should still return False to indicate that the parameters
        # are the same.
        remote = [
            {'ParameterKey': 'BucketName', 'ParameterValue': 'name'},
            {'ParameterKey': 'BucketPassword', 'ParameterValue': '***'},
            {'ParameterKey': 'DefaultParam',
             'ParameterValue': 'DefaultValue'},
            {'ParameterKey': 'Metadata', 'ParameterValue': '1.0'}
        ]
        ret = self.actor._diff_params_safely(remote, self.actor._parameters)
        self.assertEqual(False, ret)

        # Now pretend that the Metadata is different ... Should return True
        # indicating that the lists are different.
        remote = [
            {'ParameterKey': 'BucketName', 'ParameterValue': 'name'},
            {'ParameterKey': 'BucketPassword', 'ParameterValue': '***'},
            {'ParameterKey': 'Metadata', 'ParameterValue': '2.0',
             'ResolvedValue': 'Resolved'}
        ]
        ret = self.actor._diff_params_safely(remote, self.actor._parameters)
        self.assertEqual(True, ret)

        # Now try updating the parameter with a default and
        # pretend the remote had a different value. Should return True.
        self.actor._options['parameters']['DefaultParam'] = 'NewValue'
        self.actor._parameters = self.actor._create_parameters(
            self.actor._options['parameters'])
        remote = [
            {'ParameterKey': 'BucketName', 'ParameterValue': 'name'},
            {'ParameterKey': 'BucketPassword', 'ParameterValue': '***'},
            {'ParameterKey': 'Metadata', 'ParameterValue': '2.0'},
            {'ParameterKey': 'DefaultParam',
             'ParameterValue': 'EntirelyDifferentValue'},
        ]
        ret = self.actor._diff_params_safely(remote, self.actor._parameters)
        self.assertEqual(True, ret)

        # Now try updating the parameter with a default and
        # pretend the remote had the default value. Should still return True.
        self.actor._options['parameters']['DefaultParam'] = 'AnotherNewValue'
        self.actor._parameters = self.actor._create_parameters(
            self.actor._options['parameters'])
        remote = [
            {'ParameterKey': 'BucketName', 'ParameterValue': 'name'},
            {'ParameterKey': 'BucketPassword', 'ParameterValue': '***'},
            {'ParameterKey': 'Metadata', 'ParameterValue': '2.0'},
            {'ParameterKey': 'DefaultParam',
             'ParameterValue': 'DefaultValue'},
        ]
        ret = self.actor._diff_params_safely(remote, self.actor._parameters)
        self.assertEqual(True, ret)

    @testing.gen_test
    def test_update_stack_in_failed_state(self):
        fake_stack = create_fake_stack('fake', 'CREATE_FAILED')
        self.actor._get_stack = mock.MagicMock(name='_get_stack')
        self.actor._get_stack.return_value = tornado_value(fake_stack)
        self.actor._delete_stack = mock.MagicMock(name='_delete_stack')
        self.actor._delete_stack.return_value = tornado_value(fake_stack)
        self.actor._create_stack = mock.MagicMock(name='_create_stack')
        self.actor._create_stack.return_value = tornado_value(fake_stack)
        yield self.actor._update_stack(fake_stack)
        self.actor._delete_stack.assert_called_with(
            stack='arn:aws:cloudformation:us-east-1:xxxx:stack/fake/x')
        self.actor._create_stack.assert_called_with(
            stack='fake')

    @testing.gen_test
    def test_update_stack_in_delete_failed_state(self):
        fake_stack = create_fake_stack('fake', 'DELETE_FAILED')
        self.actor._get_stack = mock.MagicMock(name='_get_stack')
        with self.assertRaises(cloudformation.StackFailed):
            yield self.actor._update_stack(fake_stack)

    @testing.gen_test
    def test_update_stack_ensure_template(self):
        fake_stack = create_fake_stack('fake', 'CREATE_COMPLETE')
        self.actor._ensure_template = mock.MagicMock(name='_ensure_stack')
        self.actor._ensure_template.return_value = tornado_value(None)
        yield self.actor._update_stack(fake_stack)
        self.actor._ensure_template.assert_called_with(fake_stack)

    @testing.gen_test
    def test_update_stack_ensure_termination_protection_default_to_true(self):
        fake_stack = create_fake_stack('fake', 'CREATE_COMPLETE')
        self.actor._options['enable_termination_protection'] = True

        self.actor._update_termination_protection = mock.MagicMock(
            name='_update_termination_protection')
        self.actor._update_termination_protection.return_value = tornado_value(
            None)

        self.actor._ensure_template = mock.MagicMock(name='_ensure_stack')
        self.actor._ensure_template.return_value = tornado_value(None)

        yield self.actor._update_stack(fake_stack)
        self.actor._update_termination_protection.assert_called_with(
            fake_stack, True)

    @testing.gen_test
    def test_update_stack_ensure_termination_protection_true_to_false(self):
        fake_stack = create_fake_stack('fake', 'CREATE_COMPLETE')
        fake_stack['EnableTerminationProtection'] = True
        self.actor._options['enable_termination_protection'] = False

        self.actor._update_termination_protection = mock.MagicMock(
            name='_update_termination_protection')
        self.actor._update_termination_protection.return_value = tornado_value(
            None)

        self.actor._ensure_template = mock.MagicMock(name='_ensure_stack')
        self.actor._ensure_template.return_value = tornado_value(None)

        yield self.actor._update_stack(fake_stack)
        self.actor._update_termination_protection.assert_called_with(
            fake_stack, False)

    @testing.gen_test
    def test_update_stack_ensure_termination_protection_true_to_true(self):
        fake_stack = create_fake_stack('fake', 'CREATE_COMPLETE')
        fake_stack['EnableTerminationProtection'] = True
        self.actor._options['enable_termination_protection'] = True

        self.actor._update_termination_protection = mock.MagicMock(
            name='_update_termination_protection')
        self.actor._update_termination_protection.return_value = tornado_value(
            None)

        self.actor._ensure_template = mock.MagicMock(name='_ensure_stack')
        self.actor._ensure_template.return_value = tornado_value(None)

        yield self.actor._update_stack(fake_stack)
        self.assertFalse(self.actor._update_termination_protection.called)

    @testing.gen_test
    def test_update_stack_update_termination_protection(self):
        fake_stack = create_fake_stack('fake', 'CREATE_COMPLETE')
        self.actor._options['enable_termination_protection'] = True

        self.actor.cf3_conn.update_termination_protection.return_value = (
            tornado_value(None))

        self.actor._ensure_template = mock.MagicMock(name='_ensure_stack')
        self.actor._ensure_template.return_value = tornado_value(None)

        yield self.actor._update_stack(fake_stack)
        self.actor.cf3_conn.update_termination_protection.assert_has_calls(
            [mock.call(
                StackName='fake',
                EnableTerminationProtection=True
            )])

    @testing.gen_test
    def test_update_stack_update_termination_protection_error(self):
        fake_stack = create_fake_stack('fake', 'CREATE_COMPLETE')
        self.actor._options['enable_termination_protection'] = True

        fake_update = {
            'ResponseMetadata': {
                'HTTPStatusCode': 400,
                'RequestId': 'dfc1a12c-22c1-11e6-80b1-8fd4cf167f54'
            },
            'Error': {
                'Message': 'Template format error: JSON not well-formed',
                'Code': 'ValidationError',
                'Type': 'Sender'
            }
        }

        self.actor.cf3_conn.update_termination_protection.side_effect = (
            ClientError(fake_update, 'FakeOperation'))
        with self.assertRaises(cloudformation.StackFailed):
            yield self.actor._update_stack(fake_stack)

    @testing.gen_test
    def test_ensure_template_with_url_quietly_exits(self):
        self.actor._template_url = 'http://fakeurl.com'
        fake_stack = create_fake_stack('fake', 'CREATE_COMPLETE')
        ret = yield self.actor._ensure_template(fake_stack)
        self.assertEqual(None, ret)

    @testing.gen_test
    def test_ensure_template_no_diff(self):
        self.actor._create_change_set = mock.MagicMock(name='_create_change')
        self.actor._wait_until_change_set_ready = mock.MagicMock(name='_wait')
        fake_stack = create_fake_stack('fake', 'CREATE_COMPLETE')

        # Grab the raw stack body from the test actor -- this is what it should
        # compare against, so this test should cause the method to bail out and
        # not make any changes.
        template = json.loads(self.actor._template_body)
        get_temp_mock = mock.MagicMock(name='_get_stack_template')
        self.actor._get_stack_template = get_temp_mock
        self.actor._get_stack_template.return_value = tornado_value(template)

        ret = yield self.actor._ensure_template(fake_stack)
        self.assertEqual(None, ret)

        self.assertFalse(self.actor._create_change_set.called)
        self.assertFalse(self.actor._wait_until_change_set_ready.called)

    @testing.gen_test
    def test_ensure_template_different(self):
        self.actor._create_change_set = mock.MagicMock(name='_create_change')
        self.actor._create_change_set.return_value = tornado_value(
            {'Id': 'abcd'})
        self.actor._wait_until_change_set_ready = mock.MagicMock(name='_wait')
        self.actor._wait_until_change_set_ready.return_value = tornado_value(
            {'Changes': []})
        self.actor._execute_change_set = mock.MagicMock(name='_execute_change')
        self.actor._execute_change_set.return_value = tornado_value(None)
        self.actor.cf3_conn.delete_change_set.return_value = tornado_value(
            None)

        # Change the actors parameters from the first time it was run -- this
        # ensures all the lines on the ensure_template method are called
        self.actor._parameters = self.actor._create_parameters({})

        fake_stack = create_fake_stack('fake', 'CREATE_COMPLETE')

        # Grab the raw stack body from the test actor -- this is what it should
        # compare against, so this test should cause the method to bail out and
        # not make any changes.
        template = {'Fake': 'Stack'}
        get_temp_mock = mock.MagicMock(name='_get_stack_template')
        self.actor._get_stack_template = get_temp_mock
        self.actor._get_stack_template.return_value = tornado_value(template)

        # We run three tests in here because the setup takes so many lines
        # (above). First test is a normal execution with changes detected.
        ret = yield self.actor._ensure_template(fake_stack)
        self.assertEqual(None, ret)
        self.actor._create_change_set.assert_has_calls(
            [mock.call(fake_stack)])
        self.actor._wait_until_change_set_ready.assert_has_calls(
            [mock.call('abcd', 'Status', 'CREATE_COMPLETE')])
        self.assertFalse(self.actor.cf3_conn.delete_change_set.called)

        # Quick second execution with _dry set. In this case, we SHOULD call
        # the delete changset function.
        self.actor._dry = True
        yield self.actor._ensure_template(fake_stack)
        self.actor.cf3_conn.delete_change_set.assert_has_calls(
            [mock.call(ChangeSetName='abcd')])

    @testing.gen_test
    def test_ensure_template_exc(self):
        self.actor._create_change_set = mock.MagicMock(name='_create_change')
        self.actor._create_change_set.return_value = tornado_value(
            {'Id': 'abcd'})
        self.actor._wait_until_change_set_ready = mock.MagicMock(name='_wait')
        self.actor._wait_until_change_set_ready.return_value = tornado_value(
            {'Changes': []})
        self.actor._execute_change_set = mock.MagicMock(name='_execute_change')
        fake_exc = {
            'ResponseMetadata': {
                'HTTPStatusCode': 400,
                'RequestId': 'dfc1a12c-22c1-11e6-80b1-8fd4cf167f54'
            },
            'Error': {
                'Message': 'Template format error: JSON not well-formed',
                'Code': 'ValidationError',
                'Type': 'Sender'
            }
        }
        self.actor._execute_change_set.side_effect = ClientError(
            fake_exc, 'FakeOperation')

        fake_stack = create_fake_stack('fake', 'CREATE_COMPLETE')

        # Grab the raw stack body from the test actor -- this is what it should
        # compare against, so this test should cause the method to bail out and
        # not make any changes.
        template = {'Fake': 'Stack'}
        get_temp_mock = mock.MagicMock(name='_get_stack_template')
        self.actor._get_stack_template = get_temp_mock
        self.actor._get_stack_template.return_value = tornado_value(template)

        # Ensure we raise an exception if something bad happens
        with self.assertRaises(cloudformation.StackFailed):
            yield self.actor._ensure_template(fake_stack)

    @testing.gen_test
    def test_create_change_set_body(self):
        self.actor.cf3_conn.create_change_set.return_value = {'Id': 'abcd'}
        fake_stack = create_fake_stack('fake', 'CREATE_COMPLETE')
        ret = yield self.actor._create_change_set(fake_stack, 'uuid')
        self.assertEqual(ret, {'Id': 'abcd'})
        self.actor.cf3_conn.create_change_set.assert_has_calls(
            [mock.call(
                StackName='arn:aws:cloudformation:us-east-1:xxxx:stack/fake/x',
                TemplateBody='{"blank": "json"}',
                Capabilities=[],
                ChangeSetName='kingpin-uuid',
                Parameters=[
                    {'ParameterValue': 'value1', 'ParameterKey': 'key1'}
                ],
                UsePreviousTemplate=False,
            )])

    @testing.gen_test
    def test_create_change_set_body_with_role(self):
        self.actor.cf3_conn.create_change_set.return_value = {'Id': 'abcd'}
        fake_stack = create_fake_stack('fake', 'CREATE_COMPLETE')
        self.actor._options['role_arn'] = 'test_role_arn'
        ret = yield self.actor._create_change_set(fake_stack, 'uuid')
        self.assertEqual(ret, {'Id': 'abcd'})
        self.actor.cf3_conn.create_change_set.assert_has_calls(
            [mock.call(
                StackName='arn:aws:cloudformation:us-east-1:xxxx:stack/fake/x',
                TemplateBody='{"blank": "json"}',
                RoleARN='test_role_arn',
                Capabilities=[],
                ChangeSetName='kingpin-uuid',
                Parameters=[
                    {'ParameterValue': 'value1', 'ParameterKey': 'key1'}
                ],
                UsePreviousTemplate=False,
            )])

    @testing.gen_test
    def test_create_change_set_url(self):
        self.actor.cf3_conn.create_change_set.return_value = {'Id': 'abcd'}
        self.actor._template_body = None
        self.actor._template_url = 'http://foobar.bin'
        fake_stack = create_fake_stack('fake', 'CREATE_COMPLETE')
        ret = yield self.actor._create_change_set(fake_stack, 'uuid')
        self.assertEqual(ret, {'Id': 'abcd'})
        self.actor.cf3_conn.create_change_set.assert_has_calls(
            [mock.call(
                StackName='arn:aws:cloudformation:us-east-1:xxxx:stack/fake/x',
                TemplateURL='http://foobar.bin',
                Capabilities=[],
                ChangeSetName='kingpin-uuid',
                Parameters=[
                    {'ParameterValue': 'value1', 'ParameterKey': 'key1'}
                ],
                UsePreviousTemplate=False,
            )])

    @testing.gen_test
    def test_create_change_set_exc(self):
        self.actor.cf3_conn.create_change_set.return_value = {'Id': 'abcd'}
        fake_exc = {
            'ResponseMetadata': {
                'HTTPStatusCode': 400,
                'RequestId': 'dfc1a12c-22c1-11e6-80b1-8fd4cf167f54'
            },
            'Error': {
                'Message': 'Template format error: JSON not well-formed',
                'Code': 'ValidationError',
                'Type': 'Sender'
            }
        }
        self.actor.cf3_conn.create_change_set.side_effect = ClientError(
            fake_exc, 'FakeOperation')
        fake_stack = create_fake_stack('fake', 'CREATE_COMPLETE')
        with self.assertRaises(cloudformation.CloudFormationError):
            yield self.actor._create_change_set(fake_stack, 'uuid')

    @testing.gen_test
    def test_wait_until_change_set_ready_complete(self):
        available = {'Status': 'AVAILABLE'}
        update_in_progress = {'Status': 'UPDATE_IN_PROGRESS'}
        update_complete = {'Status': 'UPDATE_COMPLETE'}
        fake_exc = {
            'ResponseMetadata': {
                'HTTPStatusCode': 400,
                'RequestId': 'dfc1a12c-22c1-11e6-80b1-8fd4cf167f54'
            },
            'Error': {
                'Message': 'Template format error: JSON not well-formed',
                'Code': 'ValidationError',
                'Type': 'Sender'
            }
        }
        self.actor.cf3_conn.describe_change_set.side_effect = [
            available,
            update_in_progress,
            update_in_progress,
            ClientError(fake_exc, 'Failure'),
            update_complete
        ]
        yield self.actor._wait_until_change_set_ready(
            'test', 'Status', 'UPDATE_COMPLETE', sleep=0.01)
        self.actor.cf3_conn.describe_change_set.assert_has_calls(
            [mock.call(ChangeSetName='test'),
             mock.call(ChangeSetName='test'),
             mock.call(ChangeSetName='test'),
             mock.call(ChangeSetName='test')])

    @testing.gen_test
    def test_wait_until_change_set_ready_failed_status(self):
        available = {'Status': 'AVAILABLE'}
        update_in_progress = {'Status': 'UPDATE_IN_PROGRESS'}
        update_failed = {
            'Status': 'UPDATE_FAILED',
            'StatusReason': 'Template error'
        }
        self.actor.cf3_conn.describe_change_set.side_effect = [
            available,
            update_in_progress,
            update_in_progress,
            update_failed
        ]
        with self.assertRaises(cloudformation.StackFailed):
            yield self.actor._wait_until_change_set_ready(
                'test', 'Status', 'UPDATE_COMPLETE', sleep=0.01)

    @testing.gen_test
    def test_wait_until_change_set_ready_failed_status_no_reason(self):
        available = {'Status': 'AVAILABLE'}
        update_in_progress = {'Status': 'UPDATE_IN_PROGRESS'}
        update_failed = {'Status': 'UPDATE_FAILED'}
        self.actor.cf3_conn.describe_change_set.side_effect = [
            available,
            update_in_progress,
            update_in_progress,
            update_failed
        ]
        with self.assertRaises(cloudformation.StackFailed):
            yield self.actor._wait_until_change_set_ready(
                'test', 'Status', 'UPDATE_COMPLETE', sleep=0.01)

    def test_print_change_set(self):
        fake_change_set = {
            'Changes': [
                {'ResourceChange':
                    {'Action': 'Created', 'ResourceType': 'S3::Bucket',
                     'LogicalResourceId': 'MyBucket',
                     'PhysicalResourceId': 'arn:123',
                     'Replacement': True}},
                {'ResourceChange':
                    {'Action': 'Created', 'ResourceType': 'S3::Bucket',
                     'LogicalResourceId': 'MySecondBucket'}}
            ]
        }
        self.actor.log = mock.MagicMock(name='log')
        self.actor._print_change_set(fake_change_set)
        self.actor.log.warning.assert_has_calls([
            mock.call(
                'Change: Created S3::Bucket MyBucket/arn:123 '
                '(Replacement? True)'),
            mock.call(
                'Change: Created S3::Bucket MySecondBucket/N/A '
                '(Replacement? False)')
        ])

    @testing.gen_test
    def test_execute_change_set(self):
        fake = {'StackId': 'arn::fake_set'}
        self.actor._wait_until_change_set_ready = mock.MagicMock(name='_wait')
        self.actor._wait_until_change_set_ready.return_value = tornado_value(
            fake)
        self.actor._wait_until_state = mock.MagicMock(name='_wait_until_state')
        self.actor._wait_until_state.return_value = tornado_value(None)

        yield self.actor._execute_change_set(change_set_name='fake_set')

        self.actor.cf3_conn.execute_change_set.assert_has_calls(
            [mock.call(ChangeSetName='fake_set')])

        self.actor._wait_until_change_set_ready.assert_has_calls(
            [mock.call('fake_set', 'ExecutionStatus', 'EXECUTE_COMPLETE')])
        self.actor._wait_until_state.assert_has_calls(
            [mock.call('arn::fake_set',
                       (cloudformation.COMPLETE +
                        cloudformation.FAILED +
                        cloudformation.DELETED))])

    @testing.gen_test
    def test_execute_change_set_exc(self):
        fake_exc = {
            'ResponseMetadata': {
                'HTTPStatusCode': 400,
                'RequestId': 'dfc1a12c-22c1-11e6-80b1-8fd4cf167f54'
            },
            'Error': {
                'Message': 'Template format error: JSON not well-formed',
                'Code': 'ValidationError',
                'Type': 'Sender'
            }
        }

        self.actor.cf3_conn.execute_change_set = mock.MagicMock(name='_wait')
        self.actor.cf3_conn.execute_change_set.side_effect = ClientError(
            fake_exc, 'FakeOperation')

        with self.assertRaises(cloudformation.StackFailed):
            yield self.actor._execute_change_set(change_set_name='fake_set')

    @testing.gen_test
    def test_ensure_stack_is_absent_and_wants_absent(self):
        self.actor._options['state'] = 'absent'
        self.actor._get_stack = mock.MagicMock(name='_get_stack')
        self.actor._get_stack.return_value = tornado_value(None)
        self.actor._delete_stack = mock.MagicMock(name='_delete_stack')
        self.actor._delete_stack.return_value = tornado_value(None)
        self.actor._create_stack = mock.MagicMock(name='_create_stack')
        self.actor._create_stack.return_value = tornado_value(None)

        yield self.actor._ensure_stack()

        self.assertFalse(self.actor._create_stack.called)
        self.assertFalse(self.actor._delete_stack.called)

    @testing.gen_test
    def test_ensure_stack_is_present_and_wants_absent(self):
        self.actor._options['state'] = 'absent'
        fake_stack = create_fake_stack('unit-test-cf', 'CREATE_COMPLETE')
        self.actor._get_stack = mock.MagicMock(name='_get_stack')
        self.actor._get_stack.return_value = tornado_value(fake_stack)
        self.actor._delete_stack = mock.MagicMock(name='_delete_stack')
        self.actor._delete_stack.return_value = tornado_value(None)
        self.actor._create_stack = mock.MagicMock(name='_create_stack')
        self.actor._create_stack.return_value = tornado_value(None)

        yield self.actor._ensure_stack()

        self.assertTrue(self.actor._delete_stack.called)
        self.assertFalse(self.actor._create_stack.called)

    @testing.gen_test
    def test_ensure_stack_is_absent_and_wants_present(self):
        self.actor._options['state'] = 'present'
        self.actor._get_stack = mock.MagicMock(name='_get_stack')
        self.actor._get_stack.return_value = tornado_value(None)
        self.actor._delete_stack = mock.MagicMock(name='_delete_stack')
        self.actor._delete_stack.return_value = tornado_value(None)
        self.actor._create_stack = mock.MagicMock(name='_create_stack')
        self.actor._create_stack.return_value = tornado_value(None)

        yield self.actor._ensure_stack()

        self.assertFalse(self.actor._delete_stack.called)
        self.assertTrue(self.actor._create_stack.called)

    @testing.gen_test
    def test_ensure_stack_is_present_and_wants_update_create_failed(self):
        self.actor._options['state'] = 'present'
        fake_stack = create_fake_stack('fake', 'CREATE_FAILED')
        self.actor._get_stack = mock.MagicMock(name='_get_stack')
        self.actor._get_stack.return_value = tornado_value(fake_stack)
        self.actor._delete_stack = mock.MagicMock(name='_delete_stack')
        self.actor._delete_stack.return_value = tornado_value(None)
        self.actor._create_stack = mock.MagicMock(name='_create_stack')
        self.actor._create_stack.return_value = tornado_value(None)
        self.actor._update_stack = mock.MagicMock(name='_update_stack')
        self.actor._update_stack.return_value = tornado_value(None)

        yield self.actor._ensure_stack()

        self.assertFalse(self.actor._delete_stack.called)
        self.assertFalse(self.actor._create_stack.called)
        self.assertTrue(self.actor._update_stack.called)

    @testing.gen_test
    def test_execute(self):
        self.actor._validate_template = mock.MagicMock()
        self.actor._validate_template.return_value = tornado_value(None)
        self.actor._ensure_stack = mock.MagicMock()
        self.actor._ensure_stack.return_value = tornado_value(None)
        yield self.actor._execute()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2018 Nextdoor.com, Inc

import jsonschema

from kingpin.actors import exceptions


__author__ = 'Mikhail Simin <mikhail@nextdoor.com>'


class REQUIRED(object):

    """Meta class to identify required arguments for actors."""


class StringCompareBase(object):

    """Meta class to identify the desired state for a resource.

    This basic type of constant allows someone to easily define a set of valid
    strings for their option and have the base actor class automatically
    validate the inputs against those strings.
    """

    valid = None

    @classmethod
    def validate(self, option):
        if option not in self.valid:
            raise exceptions.InvalidOptions(
                '%s not valid, use: %s' % (option, self.valid))


class STATE(StringCompareBase):

    """Meta class to identify the desired state for a resource.

    Simple tester for 'present' or 'absent' on actors. Used for any actor thats
    idempotent and used to ensure some state of a resource.
    """

    valid = ('present', 'absent')


class SchemaCompareBase(object):

    """Meta class that compares the schema of a dict against rules."""

    SCHEMA = None

    @classmethod
    def validate(self, option):
        try:
            jsonschema.Draft4Validator(self.SCHEMA).validate(option)
        except jsonschema.exceptions.ValidationError as e:
            raise exceptions.InvalidOptions(
                'Supplied parameter does not match schema: %s' % e)



import demjson
import os

import unittest

from kingpin import exceptions
from kingpin import schema


class TestSchema(unittest.TestCase):

    def setUp(self, *args, **kwargs):
        super(TestSchema, self).setUp(*args, **kwargs)

        dirname, filename = os.path.split(os.path.abspath(__file__))
        self.examples = '%s/../../examples' % dirname

    def test_validate_with_simple_json(self):
        json = demjson.decode(open('%s/simple.json' % self.examples).read())
        ret = schema.validate(json)
        self.assertEqual(None, ret)

    def test_validate_with_complex_json(self):
        json = demjson.decode(open('%s/complex.json' % self.examples).read())
        ret = schema.validate(json)
        self.assertEqual(None, ret)

    def test_validate_with_invalid_json(self):
        json = {'this': 'is', 'invalid': 'ok'}
        with self.assertRaises(exceptions.InvalidScript):
            schema.validate(json)

    def test_validate_with_array_syntax(self):
        json = [{'actor': 'some actor'}]
        schema.validate(json)

    def test_validate_with_invalid_array(self):
        json = [{'garbage': 'json'}]
        with self.assertRaises(exceptions.InvalidScript):
            schema.validate(json)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2018 Nextdoor.com, Inc

"""
:mod:`kingpin.actors.misc`
^^^^^^^^^^^^^^^^^^^^^^^^^^

These are common utility Actors that don't really need their own
dedicated packages. Things like sleep timers, loggers, etc.

**Optional Environment Variables**

:URLLIB_DEBUG:
  Set this variable to enable extreme debug logging of the URLLIB requests made
  by the RightScale/AWS actors.  *Note, this is very insecure as
  headers/cookies/etc. are exposed*
"""

import io
import json
import logging
import urllib.request
import urllib.parse
import urllib.error

from tornado import gen
from tornado import httpclient
from kingpin.actors import utils as actor_utils
from kingpin.actors import group
from kingpin import exceptions as kingpin_exceptions

from kingpin import schema
from kingpin import utils
from kingpin.actors import base
from kingpin.actors import exceptions
from kingpin.constants import REQUIRED

log = logging.getLogger(__name__)

__author__ = ('Matt Wise <matt@nextdoor.com>, '
              'Mikhail Simin <mikhail@nextdoor.com>')


class Note(base.BaseActor):

    """Print any message to log."""

    all_options = {
        'message': (str, REQUIRED, 'Message to log.')
    }

    desc = "Info Log"

    @gen.coroutine
    def _execute(self):
        self.log.info(self.option('message'))


class Macro(base.BaseActor):

    """Parses a kingpin script, instantiates and executes it.

    **Parse JSON/YAML**

    Kingpin JSON/YAML has 2 passes at its validity. Script syntax must be
    valid, with the exception of a few useful deviations allowed by `demjson
    <http://deron.meranda.us/python/demjson/>`_ parser. Main
    one being the permission of inline comments via ``/* this */`` syntax.

    The second pass is validating the Schema. The script will be validated
    for schema-conformity as one of the first things that happens at load-time
    when the app starts up. If it fails, you will be notified immediately.

    Lastly after the JSON/YAML is established to be valid, all the tokens are
    replaced with their specified value. Any key/value pair passed in the
    ``tokens`` option will be available inside of the JSON file as ``%KEY%``
    and replaced with the value at this time.

    In a situation where nested Macro executions are invoked the tokens *do
    not* propagate from outter macro into the inner. This allows to reuse token
    names, but forces the user to specify every token needed. Similarly, if
    environment variables are used for token replacement in the main file,
    these tokens are not available in the subsequent macros.

    **Pre-Instantiation**

    In an effort to prevent mid-run errors, we pre-instantiate all Actor
    objects all at once before we ever begin executing code. This ensures that
    major typos or misconfigurations in the JSON/YAML will be caught early on.

    **Execution**

    `misc.Macro` actor simply calls the `execute()` method of the most-outter
    actor; be it a single action, or a group actor.

    **Options**

    :macro:
      String of local path to a JSON/YAML script.

    :tokens:
      Dictionary to search/replace within the file.

    **Examples**

    .. code-block:: json

       { "desc": "Stage 1",
         "actor": "misc.Macro",
         "options": {
           "macro": "deployment/stage-1.json",
           "tokens": {
             "TIMEOUT": 360,
             "RELEASE": "%RELEASE%"
           }
         }
       }

    **Dry Mode**

    Fully supported -- instantiates the actor inside of JSON with dry=True. The
    behavior of the consecutive actor is unique to each; read their description
    for more information on dry mode.
    """

    # By default, group actors have no timeout. We rely on the individual
    # actors to expire on their own. This is, of course, overrideable in the
    # JSON.
    default_timeout = None

    all_options = {
        'macro': (str, REQUIRED,
                  "Path to a Kingpin script. http(s)://, file:///, "
                  "absolute or relative file paths."),
        'tokens': (dict, {}, "Tokens passed into the JSON file.")
    }

    desc = "Macro: {macro}"

    def __init__(self, *args, **kwargs):
        """Pre-parse the script file and compile actors.

        Note, we override the default init_tokens={} from the base class and
        default it to a _copy_ of the os.environ dict.
        """
        super(Macro, self).__init__(*args, **kwargs)

        # Temporary check that macro is a local file.
        self._check_macro()

        self.log.info('Preparing actors from %s' % self.option('macro'))

        # Take the "init tokens" that were supplied to this actor by its parent
        # and merge them with the explicitly defined tokens in the actor
        # definition itself. Give priority to the explicitly defined tokens on
        # any conflicts.
        self._init_tokens.update(self.option('tokens'))

        # Copy the tmp file / download a remote macro
        macro_file = self._get_macro()

        # Parse script, and insert tokens.
        config = self._get_config_from_script(macro_file)

        # Check schema for compatibility
        self._check_schema(config)

        # Instantiate the first actor, but don't execute it.
        # Any errors raised by this actor should be attributed to it, and not
        # this Macro actor. No try/catch here
        if type(config) == list:
            # List is a Sync group actor
            self.initial_actor = group.Sync(options={'acts': config},
                                            dry=self._dry)
        else:
            # After the schema has been checked, pass in whatever tokens _we_
            # got, off to the soon-to-be-created actor.
            config['init_tokens'] = self._init_tokens.copy()

            self.initial_actor = actor_utils.get_actor(config, dry=self._dry)

    def _check_macro(self):
        """For now we are limiting the functionality."""

        prohibited = ('ftp://',)
        if self.option('macro').startswith(prohibited):
            raise exceptions.UnrecoverableActorFailure(
                'Macro actor is cannot handle ftp fetching yet..')

    def _get_macro(self):
        """Return a buffer to the macro file.

        Will download a remote file in-memory and return a buffer, or
        open the local file and return a buffer to that file.
        """

        remote = ('http://', 'https://')
        if self.option('macro').startswith(remote):
            client = httpclient.HTTPClient()
            try:
                R = client.fetch(self.option('macro'))
            except Exception as e:
                raise exceptions.UnrecoverableActorFailure(e)
            finally:
                client.close()
            buf = io.StringIO()
            # Set buffer representation for debug printing.
            buf.__repr__ = lambda: (
                'In-memory file from: %s' % self.option('macro'))
            buf.write(R.body)
            buf.seek(0)
            client.close()
            return buf

        try:
            instance = open(self.option('macro'))
        except IOError as e:
            raise exceptions.UnrecoverableActorFailure(e)
        return instance

    def _get_config_from_script(self, script_file):
        """Convert a script into a dict() with inserted ENV vars.

        Run the JSON dictionary through our environment parser and return
        back a dictionary with all of the %XX% keys swapped out with
        environment variables.

        Args:
            script_file: A path string to a file, or an open() file stream.

        Returns:
            Dictionary adhering to our schema.

        Raises:
            UnrecoverableActorFailure -
                if parsing script or inserting env vars fails.
        """
        self.log.debug('Parsing %s' % script_file)
        try:
            return utils.convert_script_to_dict(
                script_file=script_file,
                tokens=self._init_tokens)
        except (kingpin_exceptions.InvalidScript, LookupError) as e:
            raise exceptions.UnrecoverableActorFailure(e)

    def _check_schema(self, config):
        # Run the dict through our schema validator quickly
        self.log.debug('Validating schema for %s' % self.option('macro'))
        try:
            schema.validate(config)
        except kingpin_exceptions.InvalidScript as e:
            self.log.critical('Invalid Schema.')
            raise exceptions.UnrecoverableActorFailure(e)

    def get_orgchart(self, parent=''):
        """Return orgchart including the actor inside of the macro file."""
        ret = super(Macro, self).get_orgchart(parent=parent)
        macro = self.initial_actor.get_orgchart(parent=str(id(self)))
        return ret + macro

    @gen.coroutine
    def _execute(self):
        # initial_actor is configured with same dry parameter as this actor.
        # Just execute it and the rest will be handled internally.
        yield self.initial_actor.execute()


class Sleep(base.BaseActor):

    """Sleeps for an arbitrary number of seconds.

    **Options**

    :sleep:
      Integer of seconds to sleep.

    **Examples**

    .. code-block:: json

       { "actor": "misc.Sleep",
         "desc": "Sleep for 60 seconds",
         "options": {
           "sleep": 60
         }
       }

    **Dry Mode**

    Fully supported -- does not actually sleep, just pretends to.
    """

    all_options = {
        'sleep': ((int, float, str), REQUIRED,
                  'Number of seconds to do nothing.')
    }

    desc = "Sleep {sleep}s"

    @gen.coroutine
    def _execute(self):
        """Executes an actor and yields the results when its finished."""

        self.log.debug('Sleeping for %s seconds' % self.option('sleep'))

        sleep = self.option('sleep')

        if isinstance(sleep, str):
            sleep = float(sleep)

        if not self._dry:
            yield utils.tornado_sleep(seconds=sleep)


class GenericHTTP(base.HTTPBaseActor):

    """A very simple actor that allows GET/POST methods over HTTP.

    Does a GET or a POST to a specified URL.

    **Options**

    :url:
      Destination URL

    :data:
      Optional POST data as a `dict`. Will convert into key=value&key2=value2..
      Exclusive of `data-json` option.

    :data-json:
      Optional POST data as a `dict`. Will stringify and pass as JSON.
      Exclusive of `data` option.

    :username:
      Optional for HTTPAuth.

    :password:
      Optional for HTTPAuth.

    **Examples**

    .. code-block:: json

       { "actor": "misc.GenericHTTP",
         "desc": "Make a simple web call",
         "options": {
           "url": "http://example.com/rest/api/v1?id=123&action=doit",
           "username": "secret",
           "password": "%SECRET_PASSWORD%"
         }
       }

    **Dry Mode**

    Will not do anything in dry mode except print a log statement.
    """

    all_options = {
        'url': (str, REQUIRED, 'Domain name + query string to fetch'),
        'data': (dict, {}, 'Data to attach as a POST query'),
        'data-json': (dict, {}, 'JSON data to attach as POST query'),
        'username': (str, '', 'HTTPAuth username'),
        'password': (str, '', 'HTTPAuth password')
    }

    @gen.coroutine
    def _execute_dry(self):
        is_post = bool(self.option('data'))
        method = ['GET', 'POST'][is_post]

        self.log.info("Would do a %s request to %s"
                      % (method, self.option('url')))
        raise gen.Return()

    @gen.coroutine
    def _execute(self):

        if self._dry:
            raise gen.Return(self._execute_dry())

        # Only generate a JSON text string if a populated dict was passed to
        # data-json.
        datajson = None
        if self.option('data-json'):
            datajson = json.dumps(self.option('data-json'))

        escaped_post = (
            urllib.parse.urlencode(self.option('data')) or
            datajson or None)

        try:
            yield self._fetch(self.option('url'),
                              post=escaped_post,
                              auth_username=self.option('username'),
                              auth_password=self.option('password'))
        except httpclient.HTTPError as e:
            if e.code == 401:
                raise exceptions.InvalidCredentials(e.message)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2018 Nextdoor.com, Inc

"""
:mod:`kingpin.actors.slack`
^^^^^^^^^^^^^^^^^^^^^^^^^^^

The Slack Actors allow you to send messages to a Slack channel at stages during
your job execution. The actor supports dry mode by validating that the
configured API Token has access to execute the methods, without actually
sending the messages.

**Required Environment Variables**

:SLACK_TOKEN:
  Slack API Token

:SLACK_NAME:
  Slack *message from* name
  (defaults to *Kingpin*)
"""

import logging
import os
import re

from tornado import gen

from tornado_rest_client import api

from kingpin.constants import REQUIRED
from kingpin.actors import base
from kingpin.actors import exceptions

log = logging.getLogger(__name__)

__author__ = 'Matt Wise <matt@nextdoor.com>'


TOKEN = os.getenv('SLACK_TOKEN', None)
NAME = os.getenv('SLACK_NAME', 'Kingpin')


class SlackAPI(api.RestConsumer):

    _ENDPOINT = 'https://api.slack.com'
    _CONFIG = {
        'attrs': {
            'auth_test': {
                'path': '/api/auth.test',
                'http_methods': {'post': {}},
            },
            'chat_postMessage': {
                'path': '/api/chat.postMessage',
                'http_methods': {'post': {}},
            }

        }
    }


class SlackBase(base.BaseActor):

    """Simple Slack Abstract Base Object"""

    def __init__(self, *args, **kwargs):
        """Check required environment variables."""
        super(SlackBase, self).__init__(*args, **kwargs)

        if not TOKEN:
            raise exceptions.InvalidCredentials(
                'Missing the "SLACK_TOKEN" environment variable.')

        rest_client = api.SimpleTokenRestClient(
            tokens={'token': TOKEN}
        )
        self._slack_client = SlackAPI(client=rest_client)

    def _check_results(self, result):
        """Returns True/False if the result was OK from Slack.

        The Slack API avoids using standard error codes, and instead embeds
        error codes in the return results. This method returns True or False
        based on those results.

        Args:
            result: A return dict from Slack

        Raises:
            InvalidCredentials if the creds are bad
            RecoverableActorException on any other value
        """
        try:
            ok = result.get('ok', False)
        except AttributeError:
            raise exceptions.UnrecoverableActorFailure(
                'An unexpected Slack API failure occured: %s' % result)

        if ok:
            return

        # By default, our exception type is a RecoverableActorFailure.
        exc = exceptions.RecoverableActorFailure

        # If we know what kind fo error it is, we'll return a more accurate
        # exception type.
        if result['error'] == 'invalid_auth':
            exc = exceptions.InvalidCredentials

        # Finally, raise our exception
        raise exc('Slack API Error: %s' % result['error'])


class Message(SlackBase):

    """Sends a message to a channel in Slack.

    **Options**

    :channel:
      The string-name of the channel to send a message to, or a list of
      channels

    :message:
      String of the message to send

    **Examples**

    .. code-block:: json

       { "desc": "Let the Engineers know things are happening",
         "actor": "slack.Message",
         "options": {
           "channel": "#operations",
           "message": "Beginning Deploy: %VER%"
         }
       }

    **Dry Mode**

    Fully supported -- does not actually send messages to a room, but validates
    that the API credentials would have access to send the message using the
    Slack `auth.test` API method.
    """

    all_options = {
        'channel': ((str, list), REQUIRED, 'Slack channel or a list of names'),
        'message': (str, REQUIRED, 'Message to send')
    }

    desc = "Sending Message to {channel}"

    @gen.coroutine
    def _execute(self):
        self.log.info('Sending message "%s" to Slack channel "%s"' %
                      (self.option('message'), self.option('channel')))

        if self._dry:
            # Check if our authentication creds are valid
            auth_ok = yield self._slack_client.auth_test().http_post()
            self._check_results(auth_ok)

            self.log.info('API Credentials verified, skipping send.')
            raise gen.Return()

        # If only one channel was supplied as string then prepare the list
        if type(self.option('channel')) == list:
            channels = self.option('channel')
        else:
            channels = re.split('[, ]+', self.option('channel'))

        posts = []
        for channel in channels:
            self.log.debug('Posting to %s' % channel)
            # Finally, send the message and check our return value
            posts.append(self._slack_client.chat_postMessage().http_post(
                channel=channel,
                text=self.option('message'),
                username=NAME,
                parse='none',
                link_names=1,
                unfurl_links=True,
                unfurl_media=True
            ))

        results = yield posts
        for res in results:
            self._check_results(res)



import logging

from tornado import testing
import mock
import botocore.exceptions

from kingpin.actors import exceptions
from kingpin.actors.aws import elbv2 as elbv2_actor
from kingpin.actors.aws import settings
from kingpin.actors.test import helper
import importlib

log = logging.getLogger(__name__)


class TestRegisterInstance(testing.AsyncTestCase):

    def setUp(self):
        super(TestRegisterInstance, self).setUp()
        settings.AWS_ACCESS_KEY_ID = 'unit-test'
        settings.AWS_SECRET_ACCESS_KEY = 'unit-test'
        settings.RETRYING_SETTINGS = {'stop_max_attempt_number': 1}
        importlib.reload(elbv2_actor)

    @testing.gen_test
    def test_add(self):
        actor = elbv2_actor.RegisterInstance('UTA', {
            'target_group': 'test',
            'region': 'us-east-1',
            'instances': 'test'})
        actor.elbv2_conn = mock.Mock()
        actor.elbv2_conn.register_targets.return_value = {}

        yield actor._add('target_group_arn', ['i-un173s7'])

        actor.elbv2_conn.register_targets.assert_called_with(
            TargetGroupArn='target_group_arn',
            Targets=[{'Id': 'i-un173s7'}])

    @testing.gen_test
    def test_add_exception(self):
        actor = elbv2_actor.RegisterInstance('UTA', {
            'target_group': 'test',
            'region': 'us-east-1',
            'instances': 'test'})
        actor.elbv2_conn = mock.Mock()
        exc = botocore.exceptions.ClientError({'Error': {'Code': ''}}, 'Test')
        actor.elbv2_conn.register_targets.side_effect = exc

        with self.assertRaises(exceptions.UnrecoverableActorFailure):
            yield actor._add('target_group_arn', ['i-un173s7'])

    @testing.gen_test
    def test_execute(self):
        actor = elbv2_actor.RegisterInstance('UTA', {
            'target_group': 'test',
            'region': 'us-east-1',
            'instances': 'i-test'})

        actor._find_target_group = mock.Mock()
        actor._find_target_group.return_value = helper.tornado_value('arn')
        actor._add = mock.Mock()
        actor._add.return_value = helper.tornado_value(mock.Mock())
        yield actor._execute()
        actor._add.assert_called_with('arn', ['i-test'])

    @testing.gen_test
    def test_execute_self(self):
        # No instance id specified
        actor = elbv2_actor.RegisterInstance('UTA', {
            'target_group': 'test',
            'region': 'us-east-1'})

        actor._find_target_group = mock.Mock()
        actor._find_target_group.return_value = helper.tornado_value('arn')
        actor._add = mock.Mock()
        actor._get_meta_data = helper.mock_tornado('i-test')
        actor._add.return_value = helper.tornado_value(mock.Mock())
        yield actor._execute()
        actor._add.assert_called_with('arn', ['i-test'])


class TestDeregisterInstance(testing.AsyncTestCase):

    def setUp(self):
        super(TestDeregisterInstance, self).setUp()
        settings.AWS_ACCESS_KEY_ID = 'unit-test'
        settings.AWS_SECRET_ACCESS_KEY = 'unit-test'
        settings.RETRYING_SETTINGS = {'stop_max_attempt_number': 1}
        importlib.reload(elbv2_actor)

    @testing.gen_test
    def test_remove(self):
        actor = elbv2_actor.DeregisterInstance('UTA', {
            'target_group': 'test',
            'region': 'us-east-1',
            'instances': 'test'})
        actor.elbv2_conn = mock.Mock()
        actor.elbv2_conn.deregister_targets.return_value = {}

        yield actor._remove('target_group_arn', ['i-un173s7'])

        actor.elbv2_conn.deregister_targets.assert_called_with(
            TargetGroupArn='target_group_arn',
            Targets=[{'Id': 'i-un173s7'}])

    @testing.gen_test
    def test_remove_exception(self):
        actor = elbv2_actor.DeregisterInstance('UTA', {
            'target_group': 'test',
            'region': 'us-east-1',
            'instances': 'test'})
        actor.elbv2_conn = mock.Mock()
        exc = botocore.exceptions.ClientError({'Error': {'Code': ''}}, 'Test')
        actor.elbv2_conn.deregister_targets.side_effect = exc

        with self.assertRaises(exceptions.UnrecoverableActorFailure):
            yield actor._remove('target_group_arn', ['i-un173s7'])

    @testing.gen_test
    def test_execute(self):
        actor = elbv2_actor.DeregisterInstance('UTA', {
            'target_group': 'elb-test',
            'region': 'us-east-1',
            'instances': 'i-test'})

        actor._find_target_group = mock.Mock()
        actor._find_target_group.return_value = helper.tornado_value('arn')
        actor._remove = mock.Mock()
        actor._remove.return_value = helper.tornado_value(mock.Mock())
        yield actor._execute()
        actor._remove.assert_called_with('arn', ['i-test'])

    @testing.gen_test
    def test_execute_self(self):
        actor = elbv2_actor.DeregisterInstance('UTA', {
            'target_group': 'elb-test',
            'region': 'us-east-1'})

        actor._find_target_group = mock.Mock()
        actor._find_target_group.return_value = helper.tornado_value('arn')
        actor._get_meta_data = helper.mock_tornado('i-test')
        actor._remove = mock.Mock()
        actor._remove.return_value = helper.tornado_value(mock.Mock())
        yield actor._execute()
        actor._remove.assert_called_with('arn', ['i-test'])



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2018 Nextdoor.com, Inc

"""
:mod:`kingpin.actors.librato`
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

The Librato Actor allows you to post an Annotation to Librato. This is
specifically useful for marking when deployments occur on your graphs for
cause/effect analysis.

**Required Environment Variables**

:LIBRATO_TOKEN:
  Librato API Token

:LIBRATO_EMAIL:
  Librato email account (i.e. username)

"""

import logging
import os
import urllib.request
import urllib.parse
import urllib.error

from tornado import gen
from tornado import httpclient

from kingpin import utils
from kingpin.actors import base
from kingpin.actors import exceptions
from kingpin.constants import REQUIRED

log = logging.getLogger(__name__)

__author__ = 'Charles McLaughlin <charles@nextdoor.com>'

API_CONTENT_TYPE = 'application/x-www-form-urlencoded'
API_URL = 'https://metrics-api.librato.com/v1/'
ANNOTATIONS_URL = API_URL + 'annotations/'
METRICS_URL = API_URL + 'metrics'  # Used to test auth

TOKEN = os.getenv('LIBRATO_TOKEN', None)
EMAIL = os.getenv('LIBRATO_EMAIL', None)


class Annotation(base.HTTPBaseActor):

    """Librato Annotation Actor

    Posts an Annotation to Librato.

    **Options**

    :title:
      The title of the annotation

    :description:
      The description of the annotation

    :name:
      Name of the metric to annotate

    **Examples**

    .. code-block:: json

       { "actor": "librato.Annotation",
         "desc": "Mark our deployment",
         "options": {
           "title": "Deploy",
           "description": "Version: 0001a",
           "name": "production_releases"
         }
       }

    **Dry Mode**

    Currently does not actually do anything, just logs dry mode.
    """

    all_options = {
        'title': (str, REQUIRED, "Annotation title"),
        'description': (str, REQUIRED, "Annotation description"),
        'name': (str, REQUIRED, "Name of the metric to annotate")
    }

    desc = "Sending Annotation to {name}"

    def __init__(self, *args, **kwargs):
        """Check for the needed environment variables."""
        super(Annotation, self).__init__(*args, **kwargs)

        if not TOKEN:
            raise exceptions.InvalidCredentials(
                'Missing the "LIBRATO_TOKEN" environment variable.')

        if not EMAIL:
            raise exceptions.InvalidCredentials(
                'Missing the "LIBRATO_EMAIL" environment variable.')

    @gen.coroutine
    @utils.retry(excs=(httpclient.HTTPError), retries=3)
    def _fetch_wrapper(self, *args, **kwargs):
        """Wrap the superclass _fetch method to catch known Librato errors."""
        try:
            res = yield self._fetch(*args, **kwargs)
        except httpclient.HTTPError as e:
            if e.code == 400:
                # "HTTPError: HTTP 400: Bad Request"
                raise exceptions.BadRequest(
                    'Check your JSON inputs.')
            if e.code == 401:
                # "The authentication you provided is invalid."
                raise exceptions.InvalidCredentials(
                    'Librato authentication failed.')
            raise

        raise gen.Return(res)

    @gen.coroutine
    def _execute(self):
        """Executes an actor and yields the results when its finished.

        raises: gen.Return()
        """

        if self._dry:
            self.log.info('Testing Librato auth, skipping annotation')
            msg = ("Would have annotated metric "
                   "'%s' with title:'%s', description:'%s'")
            self.log.info(msg % (self.option('name'), self.option('title'),
                                 self.option('description')))
            yield self._fetch_wrapper(
                METRICS_URL, auth_username=EMAIL, auth_password=TOKEN)
        else:
            self.log.info(
                "Annotating metric '%s' with title:'%s', description:'%s'" % (
                    self.option('name'), self.option('title'),
                    self.option('description')))
            url = ANNOTATIONS_URL + self.option('name')
            args = urllib.parse.urlencode(
                {'title': self.option('title'),
                 'description': self.option('description')})

            yield self._fetch_wrapper(url, post=args,
                                      auth_username=EMAIL, auth_password=TOKEN)
        raise gen.Return()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2018 Nextdoor.com, Inc

"""
:mod:`kingpin.actors.aws.base`
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

The AWS Actors allow you to interact with the resources (such as SQS and ELB)
inside your Amazon AWS account. These actors all support dry runs properly, but
each actor has its own caveats with ``dry=True``. Please read the instructions
below for using each actor.

**Required Environment Variables**

_Note, these can be skipped only if you have a .aws/credentials file in place._

:AWS_ACCESS_KEY_ID:
  Your AWS access key

:AWS_SECRET_ACCESS_KEY:
  Your AWS secret
"""

import json
import logging
import urllib.request
import urllib.parse
import urllib.error
import re

from boto import exception as boto_exception
from boto import utils as boto_utils
from boto3 import exceptions as boto3_exceptions
from botocore import exceptions as botocore_exceptions
from retrying import retry
from tornado import concurrent
from tornado import gen
from tornado import ioloop
import boto.cloudformation
import boto.ec2
import boto.ec2.elb
import boto.iam
import boto.sqs
import boto3

from kingpin import utils
from kingpin import exceptions as kingpin_exceptions
from kingpin.actors import base
from kingpin.actors import exceptions
from kingpin.actors.aws import api_call_queue
from kingpin.actors.aws import settings as aws_settings

log = logging.getLogger(__name__)

__author__ = 'Mikhail Simin <mikhail@nextdoor.com>'

EXECUTOR = concurrent.futures.ThreadPoolExecutor(10)

NAMED_API_CALL_QUEUES = {}


class ELBNotFound(exceptions.RecoverableActorFailure):
    """Raised when an ELB is not found"""


class InvalidMetaData(exceptions.UnrecoverableActorFailure):
    """Raised when fetching AWS metadata."""


class InvalidPolicy(exceptions.RecoverableActorFailure):
    """Raised when Amazon indicates that policy JSON is invalid."""


class AWSBaseActor(base.BaseActor):
    # Get references to existing objects that are used by the
    # tornado.concurrent.run_on_executor() decorator.
    ioloop = ioloop.IOLoop.current()
    executor = EXECUTOR

    all_options = {
        'region': (str, None, 'AWS Region (or zone) to connect to.')
    }

    def __init__(self, *args, **kwargs):
        """Check for required settings."""

        super(AWSBaseActor, self).__init__(*args, **kwargs)

        # By default, we will try to let Boto handle discovering its
        # credentials at instantiation time. This _can_ result in synchronous
        # API calls to the Metadata service, but those should be fast.
        key = None
        secret = None

        # In the event though that someone has explicitly set the AWS access
        # keys in the environment (either for the purposes of a unit test, or
        # because they wanted to), we use those values.
        if (aws_settings.AWS_ACCESS_KEY_ID and
                aws_settings.AWS_SECRET_ACCESS_KEY):
            key = aws_settings.AWS_ACCESS_KEY_ID
            secret = aws_settings.AWS_SECRET_ACCESS_KEY

        # On our first simple IAM connection, test the credentials and make
        # sure things worked!
        try:
            # Establish connection objects that don't require a region
            self.iam_conn = boto.iam.connection.IAMConnection(
                aws_access_key_id=key,
                aws_secret_access_key=secret)
        except boto.exception.NoAuthHandlerFound:
            raise exceptions.InvalidCredentials(
                'AWS settings imported but not all credentials are supplied. '
                'AWS_ACCESS_KEY_ID: %s, AWS_SECRET_ACCESS_KEY: %s' % (
                    aws_settings.AWS_ACCESS_KEY_ID,
                    aws_settings.AWS_SECRET_ACCESS_KEY))

        # Establish region-specific connection objects.
        region = self.option('region')
        if not region:
            return

        # In case a zone was provided instead of region we can convert
        # it on the fly
        zone_check = re.match(r'(.*[0-9])([a-z]*)$', region)

        if zone_check and zone_check.group(2):
            zone = region  # Only saving this for the log below

            # Set the fixed region
            region = zone_check.group(1)
            self.log.warning('Converting zone "%s" to region "%s".' % (
                zone, region))

        region_names = [r.name for r in boto.ec2.elb.regions()]
        if region not in region_names:
            err = ('Region "%s" not found. Available regions: %s' %
                   (region, region_names))
            raise exceptions.InvalidOptions(err)

        self.ec2_conn = boto.ec2.connect_to_region(
            region,
            aws_access_key_id=key,
            aws_secret_access_key=secret)
        self.ecs_conn = boto3.client(
            'ecs',
            region_name=region,
            aws_access_key_id=key,
            aws_secret_access_key=secret)
        self.elb_conn = boto.ec2.elb.connect_to_region(
            region,
            aws_access_key_id=key,
            aws_secret_access_key=secret)
        self.elbv2_conn = boto3.client(
            'elbv2',
            region_name=region,
            aws_access_key_id=key,
            aws_secret_access_key=secret)
        self.cf3_conn = boto3.client(
            'cloudformation',
            region_name=region,
            aws_access_key_id=key,
            aws_secret_access_key=secret)
        self.sqs_conn = boto.sqs.connect_to_region(
            region,
            aws_access_key_id=key,
            aws_secret_access_key=secret)
        self.s3_conn = boto3.client(
            's3',
            region_name=region,
            aws_access_key_id=key,
            aws_secret_access_key=secret)

    @concurrent.run_on_executor
    @retry(**aws_settings.RETRYING_SETTINGS)
    @utils.exception_logger
    def api_call(self, api_function, *args, **kwargs):
        """Execute `api_function` in a concurrent thread.

        Example:
            >>> zones = yield thread(ec2_conn.get_all_zones)
        This allows execution of any function in a thread without having
        to write a wrapper method that is decorated with run_on_executor()
        """
        try:
            return api_function(*args, **kwargs)
        except (boto_exception.BotoServerError,
                boto3_exceptions.Boto3Error) as e:
            raise self._wrap_boto_exception(e)

    @gen.coroutine
    @utils.exception_logger
    def api_call_with_queueing(self, api_function,
                               queue_name, *args, **kwargs):
        """
        Execute `api_function` in a serialized queue.

        Concurrent calls to this function are serialized into a queue.
        When any api function hits rate throttling, it backs up exponentially.

        The retry loop will always have a pause between sequential calls,
        and the delay between the calls will increase as
        recoverable api failures happen.

        The api function is assumed to be a synchronous function.
        It will be run on a concurrent thread using run_on_executor.

        The queue_identifier argument specifies which queue to use.
        If the queue doesn't exist, it will be created.

        Example:
            >>> zones = yield api_call_with_queueing(
            >>>     ec2_conn.get_all_zones, queue_name='get_all_zones')
        """
        if queue_name not in NAMED_API_CALL_QUEUES:
            NAMED_API_CALL_QUEUES[queue_name] = (
                api_call_queue.ApiCallQueue())
        queue = NAMED_API_CALL_QUEUES[queue_name]
        try:
            result = yield queue.call(api_function, *args, **kwargs)
        except (boto_exception.BotoServerError,
                boto3_exceptions.Boto3Error) as e:
            raise self._wrap_boto_exception(e)
        else:
            raise gen.Return(result)

    def _wrap_boto_exception(self, e):
        if isinstance(e, boto_exception.BotoServerError):
            # If we're using temporary IAM credentials, when those expire we
            # can get back a blank 400 from Amazon. This is confusing, but it
            # happens because of https://github.com/boto/boto/issues/898. In
            # most cases, these temporary IAM creds can be re-loaded by
            # reaching out to the AWS API (for example, if we're using an IAM
            # Instance Profile role), so thats what Boto tries to do. However,
            # if you're using short-term creds (say from SAML auth'd logins),
            # then this fails and Boto returns a blank 400.
            if (e.status == 400 and
                    e.reason == 'Bad Request' and
                    e.error_code is None):
                msg = 'Access credentials have expired'
                return exceptions.InvalidCredentials(msg)

            msg = '%s: %s' % (e.error_code, str(e))
            if e.status == 403:
                return exceptions.InvalidCredentials(msg)
        elif isinstance(e, boto3_exceptions.Boto3Error):
            return exceptions.RecoverableActorFailure(
                'Boto3 had a failure: %s' % e)
        return e

    @gen.coroutine
    def _find_elb(self, name):
        """Return an ELB with the matching name.

        Must find exactly 1 match. Zones are limited by the AWS credentials.

        Args:
            name: String-name of the ELB to search for

        Returns:
            A single ELB reference object

        Raises:
            ELBNotFound
        """
        self.log.info('Searching for ELB "%s"' % name)

        try:
            elbs = yield self.api_call(self.elb_conn.get_all_load_balancers,
                                       load_balancer_names=name)
        except boto_exception.BotoServerError as e:
            msg = '%s: %s' % (e.error_code, str(e))
            log.error('Received exception: %s' % msg)

            if e.status == 400:
                raise ELBNotFound(msg)

            raise

        self.log.debug('ELBs found: %s' % elbs)

        if len(elbs) != 1:
            raise ELBNotFound('Expected to find exactly 1 ELB. Found %s: %s'
                              % (len(elbs), elbs))

        raise gen.Return(elbs[0])

    @gen.coroutine
    def _find_target_group(self, arn):
        """Returns an ELBv2 Target Group with the matching name.

        Args:
            name: String-name of the Target Group to search for

        Returns:
            A single Target Group reference object

        Raises:
            ELBNotFound
        """
        self.log.info('Searching for Target Group "%s"' % arn)

        try:
            trgts = yield self.api_call(self.elbv2_conn.describe_target_groups,
                                        Names=[arn])
        except botocore_exceptions.ClientError as e:
            raise exceptions.UnrecoverableActorFailure(str(e))

        arns = [t['TargetGroupArn'] for t in trgts['TargetGroups']]

        if len(arns) != 1:
            raise ELBNotFound(
                'Expected to find exactly 1 Target Group. Found %s: %s'
                % (len(arns), arns))

        raise gen.Return(arns[0])

    @gen.coroutine
    def _get_meta_data(self, key):
        """Get AWS meta data for current instance.

        http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/
        ec2-instance-metadata.html
        """

        meta = yield self.api_call(boto_utils.get_instance_metadata,
                                   timeout=1, num_retries=2)
        if not meta:
            raise InvalidMetaData('No metadata available. Not AWS instance?')

        data = meta.get(key, None)
        if not data:
            raise InvalidMetaData('Metadata for key `%s` is not available')

        raise gen.Return(data)

    def _policy_doc_to_dict(self, policy):
        """Converts a Boto UUEncoded Policy document to a Dict.

        args:
            policy: The policy string returned by Boto
        """
        return json.loads(urllib.parse.unquote(policy))

    def _parse_policy_json(self, policy):
        """Parse a single JSON file into an Amazon policy.

        Validates that the policy document can be parsed, strips out any
        comments, and fills in any environmental tokens. Returns a dictionary
        of the contents.

        Returns None if the input is None.

        args:
            policy: The Policy JSON file to read.

        returns:
            A dictionary of the parsed policy.
        """
        if policy is None:
            return None

        # Run through any supplied Inline IAM Policies and verify that they're
        # not corrupt very early on.
        self.log.debug('Parsing and validating %s' % policy)

        try:
            p_doc = utils.convert_script_to_dict(script_file=policy,
                                                 tokens=self._init_tokens)
        except kingpin_exceptions.InvalidScript as e:
            raise exceptions.UnrecoverableActorFailure('Error parsing %s: %s' %
                                                       (policy, e))

        return p_doc


class EnsurableAWSBaseActor(AWSBaseActor, base.EnsurableBaseActor):

    """Ensurable version of the AWS Base Actor"""



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Copyright 2018 Nextdoor.com, Inc

"""
:mod:`kingpin.actors.aws.iam.base`
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
"""

import logging

from tornado import concurrent

from kingpin.actors.aws import base

log = logging.getLogger(__name__)

__author__ = 'Mikhail Simin <mikhail@nextdoor.com>'


# This executor is used by the tornado.concurrent.run_on_executor()
# decorator. We would like this to be a class variable so its shared
# across RightScale objects, but we see testing IO errors when we
# do this.
EXECUTOR = concurrent.futures.ThreadPoolExecutor(10)


class IAMBaseActor(base.AWSBaseActor):

    """Base class for IAM actors."""



"""Tests for the pingdom actors"""

import mock

from tornado import testing

from kingpin.actors import pingdom
from kingpin.actors import exceptions
from kingpin.actors.test.helper import mock_tornado, tornado_value


__author__ = 'Mikhail Simin <mikhail@nextdoor.com>'


class TestPingdomBase(testing.AsyncTestCase):

    def setUp(self, *args, **kwargs):
        super(TestPingdomBase, self).setUp()
        pingdom.TOKEN = 'Unittest'
        pingdom.USER = 'Unittest'
        pingdom.PASS = 'Unittest'

    @testing.gen_test
    def test_check_name(self):
        actor = pingdom.PingdomBase('Unit Test Action', {'name': 'lollipop'})

        actor._pingdom_client = mock.Mock()
        actor._pingdom_client.checks().http_get = mock_tornado(
            {'checks': [{'name': 'lollipop'}]})

        check = yield actor._get_check()

        self.assertEqual(check['name'], 'lollipop')

    @testing.gen_test
    def test_check_name_fail(self):
        actor = pingdom.PingdomBase('Unit Test Action', {'name': 'lollipop'})

        actor._pingdom_client = mock.Mock()
        actor._pingdom_client.checks().http_get = mock_tornado(
            {'checks': []})

        with self.assertRaises(exceptions.InvalidOptions):
            yield actor._get_check()


class TestPause(testing.AsyncTestCase):

    def setUp(self, *args, **kwargs):
        super(TestPause, self).setUp()
        pingdom.TOKEN = 'Unittest'
        pingdom.USER = 'Unittest'
        pingdom.PASS = 'Unittest'

    @testing.gen_test
    def test_execute(self):
        actor = pingdom.Pause('Unit Test Action', {'name': 'lollipop'})

        actor._pingdom_client = mock.Mock()
        actor._get_check = mock_tornado({'name': 'lollipop',
                                         'hostname': 'http://lollipop.com',
                                         'id': 'lol'})
        actor._pingdom_client.check().http_put.return_value = tornado_value()

        yield actor._execute()

        self.assertEqual(actor._get_check._call_count, 1)
        actor._pingdom_client.check.assert_called_with(check_id='lol')
        actor._pingdom_client.check().http_put.assert_called_with(
            paused='true')

    @testing.gen_test
    def test_execute_dry(self):
        actor = pingdom.Pause('Unit Test Action', {'name': 'lollipop'},
                              dry=True)

        actor._pingdom_client = mock.Mock()
        actor._get_check = mock_tornado({'name': 'lollipop',
                                         'hostname': 'http://lollipop.com',
                                         'id': 'lol'})

        yield actor._execute()

        self.assertEqual(actor._get_check._call_count, 1)
        actor._pingdom_client.check().http_put.assert_not_called()


class TestUnpause(testing.AsyncTestCase):

    def setUp(self, *args, **kwargs):
        super(TestUnpause, self).setUp()
        pingdom.TOKEN = 'Unittest'
        pingdom.USER = 'Unittest'
        pingdom.PASS = 'Unittest'

    @testing.gen_test
    def test_execute(self):
        actor = pingdom.Unpause('Unit Test Action', {'name': 'lollipop'})

        actor._pingdom_client = mock.Mock()
        actor._get_check = mock_tornado({'name': 'lollipop',
                                         'hostname': 'http://lollipop.com',
                                         'id': 'lol'})
        actor._pingdom_client.check().http_put.return_value = tornado_value()

        yield actor._execute()

        self.assertEqual(actor._get_check._call_count, 1)
        actor._pingdom_client.check.assert_called_with(check_id='lol')
        actor._pingdom_client.check().http_put.assert_called_with(
            paused='false')

    @testing.gen_test
    def test_execute_dry(self):
        actor = pingdom.Unpause('Unit Test Action', {'name': 'lollipop'},
                                dry=True)

        actor._pingdom_client = mock.Mock()
        actor._get_check = mock_tornado({'name': 'lollipop',
                                         'hostname': 'http://lollipop.com',
                                         'id': 'lol'})

        yield actor._execute()

        self.assertEqual(actor._get_check._call_count, 1)
        actor._pingdom_client.check().http_put.assert_not_called()



import logging

from tornado import testing
import boto.sqs.connection
import boto.sqs.queue
import mock

from kingpin.actors import exceptions
from kingpin.actors.aws import settings
from kingpin.actors.aws import sqs
from kingpin.actors.test.helper import mock_tornado
import importlib

log = logging.getLogger(__name__)


class SQSTestCase(testing.AsyncTestCase):

    def setUp(self):
        super(SQSTestCase, self).setUp()
        settings.AWS_ACCESS_KEY_ID = 'unit-test'
        settings.AWS_SECRET_ACCESS_KEY = 'unit-test'
        settings.RETRYING_SETTINGS = {'stop_max_attempt_number': 1}
        importlib.reload(sqs)

    @mock.patch.object(boto.sqs.connection, 'SQSConnection')
    def run(self, result, sqsc):
        self.sqs_conn = sqsc
        super(SQSTestCase, self).run(result=result)


class TestSQSBaseActor(SQSTestCase):

    @testing.gen_test
    def test_fetch(self):

        all_queues = [mock.Mock(), mock.Mock(), mock.Mock(), mock.Mock()]
        all_queues[0].name = '1-miss'
        all_queues[1].name = '2-miss'
        all_queues[2].name = '3-match'
        all_queues[3].name = '4-match'

        self.sqs_conn().get_all_queues.return_value = all_queues

        actor = sqs.SQSBaseActor('Unit Test Action', {
            'name': 'unit-test-queue',
            'region': 'us-east-1'})

        results = yield actor._fetch_queues('match')

        self.assertEqual(results, [all_queues[2], all_queues[3]])


class TestCreateSQSQueueActor(SQSTestCase):

    @testing.gen_test
    def test_execute(self):
        self.actor = sqs.Create('Unit Test Action',
                                {'name': 'unit-test-queue',
                                 'region': 'us-west-2'})

        self.sqs_conn().create_queue.return_value = boto.sqs.queue.Queue()
        ret = yield self.actor.execute()
        self.assertEqual(ret, None)
        self.sqs_conn().create_queue.assert_called_once_with('unit-test-queue')

    @testing.gen_test
    def test_execute_dry(self):
        self.actor = sqs.Create('Unit Test Action',
                                {'name': 'unit-test-queue',
                                 'region': 'us-west-2'},
                                dry=True)

        self.sqs_conn().create_queue.return_value = boto.sqs.queue.Queue()
        yield self.actor.execute()
        self.assertFalse(self.sqs_conn().create_queue.called)

    @testing.gen_test
    def test_execute_with_error(self):
        self.actor = sqs.Create('Unit Test Action',
                                {'name': 'unit-test-queue',
                                 'region': 'us-west-2'})

        self.sqs_conn().create_queue.return_value = False

        with self.assertRaises(exceptions.UnrecoverableActorFailure):
            yield self.actor.execute()
        self.sqs_conn().create_queue.assert_called_once_with('unit-test-queue')


class TestDeleteSQSQueueActor(SQSTestCase):

    @testing.gen_test
    def test_delete_queue(self):
        actor = sqs.Delete('Unit Test Action',
                           {'name': 'unit-test-queue',
                            'region': 'us-west-2'})
        q = mock.Mock()
        q .name = 'unit-test-queue'
        self.sqs_conn().delete_queue.return_value = False
        with self.assertRaises(sqs.QueueDeletionFailed):
            yield actor._delete_queue(queue=q)

    @testing.gen_test
    def test_execute(self):
        actor = sqs.Delete('Unit Test Action',
                           {'name': 'unit-test-queue',
                            'region': 'us-west-2'})

        q = mock.Mock()
        q.name = 'unit-test-queue'
        self.sqs_conn().get_all_queues = mock.Mock(return_value=[q])
        self.sqs_conn().delete_queue.return_value = True
        yield actor.execute()

        self.assertTrue(self.sqs_conn().get_all_queues.called)
        self.assertTrue(self.sqs_conn().delete_queue.called)

    @testing.gen_test
    def test_execute_dry(self):
        actor = sqs.Delete('Unit Test Action',
                           {'name': 'unit-test-queue',
                            'region': 'us-west-2'},
                           dry=True)

        q = mock.Mock()
        q.name = 'unit-test-queue'
        self.sqs_conn().get_all_queues = mock.Mock(return_value=[q])
        self.sqs_conn().delete_queue.return_value = True
        yield actor.execute()

        self.assertTrue(self.sqs_conn().get_all_queues.called)
        self.assertFalse(self.sqs_conn().delete_queue.called)

        self.sqs_conn().get_all_queues = mock.Mock(return_value=[])
        # Should fail even in dry run, if idempotent flag is not there.
        settings.SQS_RETRY_DELAY = 0
        importlib.reload(sqs)
        with self.assertRaises(sqs.QueueNotFound):
            yield actor.execute()

    @testing.gen_test
    def test_execute_with_failure(self):
        settings.SQS_RETRY_DELAY = 0
        importlib.reload(sqs)
        actor = sqs.Delete('Unit Test Action',
                           {'name': 'non-existent-queue',
                            'region': 'us-west-2'})

        with self.assertRaises(sqs.QueueNotFound):
            yield actor.execute()

    @testing.gen_test
    def test_execute_idempotent(self):
        settings.SQS_RETRY_DELAY = 0
        importlib.reload(sqs)
        actor = sqs.Delete('Unit Test Action',
                           {'name': 'non-existent-queue',
                            'region': 'us-west-2',
                            'idempotent': True})

        # Should work w/out raising an exception.
        yield actor.execute()


class TestWaitUntilQueueEmptyActor(SQSTestCase):

    @testing.gen_test
    def test_execute(self):
        actor = sqs.WaitUntilEmpty('UTA!',
                                   {'name': 'unit-test-queue',
                                    'region': 'us-west-2'})

        actor._wait = mock_tornado(True)
        actor._fetch_queues = mock_tornado([mock.Mock()])
        yield actor.execute()

    @testing.gen_test
    def test_execute_empty(self):
        actor = sqs.WaitUntilEmpty('UTA!',
                                   {'name': 'unit-test-queue',
                                    'region': 'us-west-2',
                                    'required': True})

        actor._wait = mock_tornado(True)
        actor._fetch_queues = mock_tornado()
        with self.assertRaises(sqs.QueueNotFound):
            yield actor.execute()

    @testing.gen_test
    def test_wait(self):
        actor = sqs.WaitUntilEmpty('UTA!',
                                   {'name': 'unit-test-queue',
                                    'region': 'us-west-2'})
        queue = mock.Mock()
        queue.count.side_effect = [1, 0, 0]
        attr = 'ApproximateNumberOfMessagesNotVisible'
        queue.get_attributes.side_effect = [{attr: '0'},
                                            {attr: '1'},
                                            {attr: '0'}]
        yield actor._wait(queue, sleep=0)
        self.assertEqual(queue.count.call_count, 3)

    @testing.gen_test
    def test_wait_dry(self):
        actor = sqs.WaitUntilEmpty('UTA!',
                                   {'name': 'unit-test-queue',
                                    'region': 'us-west-2'},
                                   dry=True)
        queue = mock.Mock()
        queue.count.side_effect = [1, 2]  # Not zero!
        yield actor._wait(queue, sleep=0)
        self.assertEqual(queue.count.call_count, 0)



"""Tests for the actors.slack package"""

from nose.plugins.attrib import attr

from tornado import testing

from kingpin.actors import slack
from kingpin.actors import exceptions
import importlib


__author__ = 'Matt Wise <matt@nextdoor.com>'


class IntegrationSlackMessage(testing.AsyncTestCase):

    """Simple high level integration tests agains the Slack API.

    These tests actually hit the Slack API and test that the code
    works, as well as validate that the API token is working properly.

    Require SLACK_TOKEN environment variable to be set.
    """

    integration = True
    message = 'Unit test message'
    channel = '#kingpin-integration'

    @attr('slack', 'integration', 'dry')
    @testing.gen_test(timeout=2)
    def integration_test_init_without_environment_creds(self):
        # Un-set the token now and make sure the init fails
        slack.TOKEN = None
        with self.assertRaises(exceptions.InvalidCredentials):
            slack.Message(
                'Unit Test Action',
                {'message': self.message,
                 'channel': self.channel}, dry=True)

        # Reload the slack library to re-get the token
        importlib.reload(slack)

    @attr('slack', 'integration', 'dry')
    @testing.gen_test(timeout=2)
    def integration_test_execute_with_invalid_creds(self):
        # Un-set the token now and make sure the init fails
        slack.TOKEN = 'unittest'
        actor = slack.Message(
            'Unit Test Action',
            {'message': self.message,
             'channel': self.channel}, dry=True)

        # Valid response test
        actor._token = 'Invalid'
        with self.assertRaises(exceptions.InvalidCredentials):
            yield actor.execute()

        # Reload the slack library to re-get the token
        importlib.reload(slack)

    @attr('slack', 'integration')
    @testing.gen_test(timeout=60)
    def integration_test_execute_invalid_room(self):
        actor = slack.Message(
            'Unit Test Action',
            {'message': self.message,
             'channel': "#bad_channel"})
        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield actor.execute()

    @attr('slack', 'integration', 'dry')
    @testing.gen_test(timeout=60)
    def integration_test_execute_dry(self):
        actor = slack.Message(
            'Unit Test Action',
            {'message': self.message,
             'channel': self.channel}, dry=True)
        res = yield actor.execute()
        self.assertEqual(res, None)

    @attr('slack', 'integration')
    @testing.gen_test(timeout=60)
    def integration_test_execute_real(self):
        actor = slack.Message(
            'Unit Test Action',
            {'message': self.message,
             'channel': self.channel})
        res = yield actor.execute()
        self.assertEqual(res, None)



import logging

from botocore.exceptions import ClientError
from tornado import testing
import mock

from kingpin.actors import exceptions
from kingpin.actors.aws import s3 as s3_actor
from kingpin.actors.aws import settings
import importlib

log = logging.getLogger(__name__)


class TestBucket(testing.AsyncTestCase):

    def setUp(self):
        super(TestBucket, self).setUp()
        settings.AWS_ACCESS_KEY_ID = 'unit-test'
        settings.AWS_SECRET_ACCESS_KEY = 'unit-test'
        settings.RETRYING_SETTINGS = {'stop_max_attempt_number': 1}
        importlib.reload(s3_actor)

        self.actor = s3_actor.Bucket(
            options={
                'name': 'test',
                'region': 'us-east-1',
                'policy': 'examples/aws.s3/amazon_put.json',
                'lifecycle': [{
                    'id': 'test',
                    'prefix': '/test',
                    'status': 'Enabled',
                    'transition': {
                        'days': 45,
                        'storage_class': 'GLACIER',
                    },
                    'noncurrent_version_transition': {
                        'noncurrent_days': 14,
                        'storage_class': 'GLACIER',
                    },
                    'expiration': '30'
                }],
                'logging': {
                    'target': 'test_target',
                    'prefix': '/prefix'
                },
                'public_access_block_configuration': {
                    'block_public_acls': True,
                    'block_public_policy': True,
                    'ignore_public_acls': True,
                    'restrict_public_buckets': True,
                },
                'versioning': False,
                'tags': [],
            })
        self.actor.s3_conn = mock.MagicMock()

    def test_init_with_bogus_logging_config(self):
        with self.assertRaises(exceptions.InvalidOptions):
            s3_actor.Bucket(
                options={
                    'name': 'test',
                    'logging': {
                        'invalid_data': 'bad_field'
                    }})

    def test_generate_lifecycle_valid_config(self):
        # Validates that the generated config called by the __init__ class is
        # correct based on the actor configuration in the setUp() method above
        self.assertEqual(len(self.actor.lifecycle), 1)

        # Verify that the rule was created with the basic options
        r = self.actor.lifecycle[0]
        self.assertEqual(r['ID'], 'test')
        self.assertEqual(r['Filter']['Prefix'], '/test')
        self.assertEqual(r['Status'], 'Enabled')

        # Validate that the string "30" was turned into an Expiration object
        self.assertEqual(r['Expiration']['Days'], 30)

        # Validate that the transition config was built properly too
        self.assertEqual(r['Transitions'][0]['Days'], 45)

        # Validate that the transition config was built properly too
        self.assertEqual(
            r['NoncurrentVersionTransitions'][0]['NoncurrentDays'], 14)

    def test_snake_to_camel(self):
        snake = {
            'i_should_be_taller': {
                'me_too_man': [
                    'not_me'
                ]
            }
        }

        self.assertEqual(
            self.actor._snake_to_camel(snake),
            {'IShouldBeTaller': {'MeTooMan': ['not_me']}}
        )

    @testing.gen_test
    def test_precache(self):
        self.actor.s3_conn.list_buckets.return_value = {
            'Buckets': [
                {'Name': 'wrong_bucket'},
                {'Name': 'test'}
            ]
        }

        yield self.actor._precache()
        self.assertTrue(self.actor._bucket_exists)

    @testing.gen_test
    def test_get_state_absent(self):
        ret = yield self.actor._get_state()
        self.assertEqual('absent', ret)

    @testing.gen_test
    def test_get_state_present(self):
        self.actor._bucket_exists = True
        ret = yield self.actor._get_state()
        self.assertEqual('present', ret)

    @testing.gen_test
    def test_set_state_absent(self):
        self.actor._options['state'] = 'absent'
        yield self.actor._set_state()
        self.actor.s3_conn.delete_bucket.assert_has_calls([
            mock.call(Bucket='test')])

    @testing.gen_test
    def test_set_state_present(self):
        self.actor._options['state'] = 'present'
        yield self.actor._set_state()
        self.actor.s3_conn.create_bucket.assert_has_calls([
            mock.call(Bucket='test')])

    @testing.gen_test
    def test_create_bucket(self):
        yield self.actor._create_bucket()
        self.actor.s3_conn.create_bucket.assert_called_with(Bucket='test')

    @testing.gen_test
    def test_create_bucket_new_region(self):
        self.actor._options['region'] = 'us-west-1'
        yield self.actor._create_bucket()
        self.actor.s3_conn.create_bucket.assert_called_with(
            Bucket='test',
            CreateBucketConfiguration={'LocationConstraint': 'us-west-1'})

    @testing.gen_test
    def test_verify_can_delete_bucket(self):
        self.actor.s3_conn.list_objects.return_value = {
            'Contents': [1, 2, 3]
        }
        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield self.actor._verify_can_delete_bucket()

    @testing.gen_test
    def test_verify_can_delete_bucket_true(self):
        self.actor.s3_conn.list_objects.return_value = {}
        yield self.actor._verify_can_delete_bucket()

    @testing.gen_test
    def test_delete_bucket(self):
        yield self.actor._delete_bucket()
        self.actor.s3_conn.delete_bucket.assert_has_calls(
            [mock.call(Bucket='test')])

    @testing.gen_test
    def test_delete_bucket_409(self):
        self.actor.s3_conn.delete_bucket.side_effect = ClientError(
            {'Error': {'Code': ''}}, 'Error')
        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield self.actor._delete_bucket()

    @testing.gen_test
    def test_get_policy(self):
        self.actor._bucket_exists = True
        self.actor.s3_conn.get_bucket_policy.return_value = {'Policy': '{}'}
        ret = yield self.actor._get_policy()
        self.assertEqual({}, ret)

    @testing.gen_test
    def test_get_policy_no_bucket(self):
        self.actor._bucket_exists = False
        ret = yield self.actor._get_policy()
        self.assertEqual(ret, None)

    @testing.gen_test
    def test_get_policy_empty(self):
        self.actor._bucket_exists = True
        self.actor.s3_conn.get_bucket_policy.side_effect = ClientError(
            {'Error': {'Code': ''}}, 'NoSuchBucketPolicy')
        ret = yield self.actor._get_policy()
        self.assertEqual('', ret)

    @testing.gen_test
    def test_get_policy_exc(self):
        self.actor._bucket_exists = True
        self.actor.s3_conn.get_bucket_policy.side_effect = ClientError(
            {'Error': {'Code': ''}}, 'SomeOtherError')
        with self.assertRaises(ClientError):
            yield self.actor._get_policy()

    @testing.gen_test
    def test_compare_policy(self):
        self.actor._bucket_exists = True
        self.actor.policy = {'test': 'policy'}
        self.actor.s3_conn.get_bucket_policy.return_value = {
            'Policy': '{"test": "policy"}'
        }
        ret = yield self.actor._compare_policy()
        self.assertTrue(ret)

    @testing.gen_test
    def test_compare_policy_false(self):
        self.actor._bucket_exists = True
        self.actor.policy = {'test': 'bah'}
        self.actor.s3_conn.get_bucket_policy.return_value = {
            'Policy': '{"test": "policy"}'
        }
        ret = yield self.actor._compare_policy()
        self.assertFalse(ret)

    @testing.gen_test
    def test_compare_policy_not_managing(self):
        self.actor._bucket_exists = True
        self.actor.policy = None
        self.actor.s3_conn.get_bucket_policy.return_value = {
            'Policy': '{"test": "policy"}'
        }
        ret = yield self.actor._compare_policy()
        self.assertTrue(ret)

    @testing.gen_test
    def test_set_policy(self):
        self.actor.policy = {}
        yield self.actor._set_policy()
        self.actor.s3_conn.put_bucket_policy.assert_has_calls([
            mock.call(Bucket='test', Policy='{}')])

    @testing.gen_test
    def test_set_policy_malformed_policy(self):
        self.actor.policy = {}
        self.actor.s3_conn.put_bucket_policy.side_effect = ClientError(
            {'Error': {'Code': ''}}, 'MalformedPolicy')
        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield self.actor._set_policy()

        self.actor.s3_conn.put_bucket_policy.assert_has_calls([
            mock.call(Bucket='test', Policy='{}')])

    @testing.gen_test
    def test_set_policy_client_error(self):
        self.actor.policy = {}
        self.actor.s3_conn.put_bucket_policy.side_effect = ClientError(
            {'Error': {'Code': ''}}, 'Some Other Error')
        with self.assertRaises(exceptions.RecoverableActorFailure):
            yield self.actor._set_policy()

        self.actor.s3_conn.put_bucket_policy.assert_has_calls([
            mock.call(Bucket='test', Policy='{}')])

    @testing.gen_test
    def test_set_policy_delete(self):
        self.actor.policy = ''
        yield self.actor._set_policy()
        self.actor.s3_conn.delete_bucket_policy.assert_has_calls([
            mock.call(Bucket='test')])

    @testing.gen_test
    def test_get_logging(self):
        self.actor._bucket_exists = True
        self.actor.s3_conn.get_bucket_logging.return_value = {
            'LoggingEnabled': {
                'TargetBucket': 'Target-Bucket',
                'TargetPrefix': 'Target-Prefix'
            }
        }

        ret = yield self.actor._get_logging()
        self.assertEqual(
            ret,
            {'target': 'Target-Bucket', 'prefix': 'Target-Prefix'})

    @testing.gen_test
    def test_get_logging_disabled(self):
        self.actor._bucket_exists = True
        self.actor.s3_conn.get_bucket_logging.return_value = {}
        ret = yield self.actor._get_logging()
        self.assertEqual(ret, {'target': '', 'prefix': ''})

    @testing.gen_test
    def test_get_logging_no_bucket(self):
        self.actor._bucket_exists = False
        ret = yield self.actor._get_logging()
        self.assertEqual(ret, None)

    @testing.gen_test
    def test_set_logging_not_desired(self):
        self.actor._options['logging'] = None
        yield self.actor._set_logging()
        self.assertFalse(self.actor.s3_conn.put_bucket_logging.called)

    @testing.gen_test
    def test_set_logging_disabled(self):
        self.actor._options['logging'] = {'target': '', 'prefix': ''}
        yield self.actor._set_logging()
        self.actor.s3_conn.put_bucket_logging.assert_has_calls([
            mock.call(Bucket='test', BucketLoggingStatus={})])

    @testing.gen_test
    def test_set_logging(self):
        self.actor._options['logging'] = {'target': 'tgt', 'prefix': 'pfx'}
        yield self.actor._set_logging()
        self.actor.s3_conn.put_bucket_logging.assert_has_calls([
            mock.call(
                Bucket='test',
                BucketLoggingStatus={'LoggingEnabled': {
                    'TargetPrefix': 'pfx', 'TargetBucket': 'tgt'}}
            )])

    @testing.gen_test
    def test_set_logging_client_error(self):
        self.actor.s3_conn.put_bucket_logging.side_effect = ClientError(
            {'Error': {'Code': ''}}, 'Some error')
        with self.assertRaises(s3_actor.InvalidBucketConfig):
            yield self.actor._set_logging()

    @testing.gen_test
    def test_get_versioning(self):
        self.actor._bucket_exists = True
        self.actor.s3_conn.get_bucket_versioning.return_value = {
            'Status': 'Enabled'}
        ret = yield self.actor._get_versioning()
        self.assertTrue(ret)

    @testing.gen_test
    def test_get_versioning_no_bucket(self):
        self.actor._bucket_exists = False
        ret = yield self.actor._get_versioning()
        self.assertEqual(None, ret)

    @testing.gen_test
    def test_get_versioning_suspended(self):
        self.actor._bucket_exists = True
        self.actor.s3_conn.get_bucket_versioning.return_value = {
            'Status': 'Suspended'}
        ret = yield self.actor._get_versioning()
        self.assertFalse(ret)

    @testing.gen_test
    def test_set_versioning(self):
        self.actor._options['versioning'] = True
        yield self.actor._set_versioning()
        self.actor.s3_conn.put_bucket_versioning.assert_has_calls([
            mock.call(
                Bucket='test',
                VersioningConfiguration={'Status': 'Enabled'})])

    @testing.gen_test
    def test_set_versioning_suspended(self):
        self.actor._options['versioning'] = False
        yield self.actor._set_versioning()
        self.actor.s3_conn.put_bucket_versioning.assert_has_calls([
            mock.call(
                Bucket='test',
                VersioningConfiguration={'Status': 'Suspended'})])

    @testing.gen_test
    def test_set_versioning_none(self):
        self.actor._options['versioning'] = None
        yield self.actor._set_versioning()
        self.assertFalse(self.actor.s3_conn.put_bucket_versioning.called)

    @testing.gen_test
    def test_get_lifecycle(self):
        self.actor._bucket_exists = True
        self.actor.s3_conn.get_bucket_lifecycle_configuration.return_value = {
            'Rules': []}
        ret = yield self.actor._get_lifecycle()
        self.assertEqual(ret, [])

    @testing.gen_test
    def test_get_lifecycle_no_bucket(self):
        self.actor._bucket_exists = False
        ret = yield self.actor._get_lifecycle()
        self.assertEqual(None, ret)

    @testing.gen_test
    def test_get_lifecycle_empty(self):
        self.actor._bucket_exists = True
        self.actor.s3_conn.get_bucket_lifecycle_configuration.side_effect = \
            ClientError(
                {'Error': {'Code': ''}}, 'NoSuchLifecycleConfiguration')
        ret = yield self.actor._get_lifecycle()
        self.assertEqual(ret, [])

    @testing.gen_test
    def test_get_lifecycle_clienterror(self):
        self.actor._bucket_exists = True
        self.actor.s3_conn.get_bucket_lifecycle_configuration.side_effect = \
            ClientError(
                {'Error': {'Code': ''}}, 'SomeOtherError')
        with self.assertRaises(ClientError):
            yield self.actor._get_lifecycle()

    @testing.gen_test
    def test_compare_lifecycle(self):
        self.actor._bucket_exists = True
        self.actor.lifecycle = [{'test': 'test'}]
        self.actor.s3_conn.get_bucket_lifecycle_configuration.return_value = {
            'Rules': [{'test': 'test'}]}
        ret = yield self.actor._compare_lifecycle()
        self.assertTrue(ret)

    @testing.gen_test
    def test_compare_lifecycle_none(self):
        self.actor.lifecycle = None
        ret = yield self.actor._compare_lifecycle()
        self.assertTrue(ret)

    @testing.gen_test
    def test_compare_lifecycle_diff(self):
        self.actor.lifecycle = [{'test1': 'test'}]
        self.actor.s3_conn.get_bucket_lifecycle_configuration.return_value = {
            'Rules': [{'test': 'test'}]}
        ret = yield self.actor._compare_lifecycle()
        self.assertFalse(ret)

    @testing.gen_test
    def test_set_lifecycle_delete(self):
        self.actor.lifecycle = []
        yield self.actor._set_lifecycle()
        self.actor.s3_conn.delete_bucket_lifecycle.assert_has_calls([
            mock.call(Bucket='test')])

    @testing.gen_test
    def test_set_lifecycle(self):
        self.actor.lifecycle = [{'test': 'test'}]
        yield self.actor._set_lifecycle()
        self.actor.s3_conn.put_bucket_lifecycle_configuration.assert_has_calls(
            [
                mock.call(
                    Bucket='test',
                    LifecycleConfiguration={'Rules': [{'test': 'test'}]})
            ]
        )

    @testing.gen_test
    def test_set_lifecycle_client_error(self):
        self.actor.s3_conn.put_bucket_lifecycle_configuration.side_effect = \
            ClientError(
                {'Error': {'Code': ''}}, 'Error')
        with self.assertRaises(s3_actor.InvalidBucketConfig):
            yield self.actor._set_lifecycle()

    @testing.gen_test
    def test_get_public_access_block_configuration(self):
        test_cfg = {
            'BlockPublicAcls': True,
            'BlockPublicPolicy': True,
            'IgnorePublicAcls': True,
            'RestrictPublicBuckets': True
        }

        self.actor._bucket_exists = True
        self.actor.s3_conn.get_public_access_block.return_value = {
            'PublicAccessBlockConfiguration': test_cfg
        }
        ret = yield self.actor._get_public_access_block_configuration()
        self.assertEqual(ret, test_cfg)

    @testing.gen_test
    def test_get_public_access_block_configuration_no_bucket(self):
        self.actor._bucket_exists = False
        ret = yield self.actor._get_public_access_block_configuration()
        self.assertEqual(None, ret)

    @testing.gen_test
    def test_get_public_access_block_configuration_empty(self):
        self.actor._bucket_exists = True
        self.actor.s3_conn.get_public_access_block.side_effect = ClientError(
            {'Error': {}}, 'NoSuchPublicAccessBlockConfiguration')
        ret = yield self.actor._get_public_access_block_configuration()
        self.assertEqual(ret, [])

    @testing.gen_test
    def test_get_public_access_block_configuration_clienterror(self):
        self.actor._bucket_exists = True
        self.actor.s3_conn.get_public_access_block.side_effect = ClientError(
            {'Error': {}}, 'SomeOtherError')
        with self.assertRaises(ClientError):
            yield self.actor._get_public_access_block_configuration()

    @testing.gen_test
    def test_compare_public_access_block_configuration(self):
        self.actor._bucket_exists = True
        self.actor.access_block = [{'test': 'test'}]
        self.actor.s3_conn.get_public_access_block.return_value = {
            'PublicAccessBlockConfiguration': [{'test': 'test'}]}
        ret = yield self.actor._compare_public_access_block_configuration()
        self.assertTrue(ret)

    @testing.gen_test
    def test_compare_public_access_block_configuration_none(self):
        self.actor.access_block = None
        ret = yield self.actor._compare_public_access_block_configuration()
        self.assertTrue(ret)

    @testing.gen_test
    def test_compare_public_access_block_configuration_diff(self):
        self.actor.access_block = [{'test1': 'test'}]
        self.actor.s3_conn.get_public_access_block.return_value = {
            'Rules': [{'test': 'test'}]}
        ret = yield self.actor._compare_public_access_block_configuration()
        self.assertFalse(ret)

    @testing.gen_test
    def test_set_public_access_block_configuration_delete(self):
        self.actor.access_block = {}
        yield self.actor._set_public_access_block_configuration()
        self.actor.s3_conn.delete_public_access_block.assert_has_calls([
            mock.call(Bucket='test')])

    @testing.gen_test
    def test_set_public_access_block_configuration(self):
        self.actor.access_block = {'test': 'test'}
        yield self.actor._set_public_access_block_configuration()
        self.actor.s3_conn.put_public_access_block.assert_has_calls([
            mock.call(
                Bucket='test',
                PublicAccessBlockConfiguration={'test': 'test'})])

    @testing.gen_test
    def test_set_public_access_block_configuration_client_error(self):
        self.actor.s3_conn.put_public_access_block.side_effect = ClientError(
            {'Error': {}}, 'Error')
        with self.assertRaises(s3_actor.InvalidBucketConfig):
            yield self.actor._set_public_access_block_configuration()

    @testing.gen_test
    def test_get_tags(self):
        self.actor._bucket_exists = True
        self.actor.s3_conn.get_bucket_tagging.return_value = {
            'TagSet': [{'Key': 'k1', 'Value': 'v1'}]}
        ret = yield self.actor._get_tags()
        self.assertEqual(ret, [{'key': 'k1', 'value': 'v1'}])

    @testing.gen_test
    def test_get_tags_multiple_tags(self):
        self.actor._bucket_exists = True
        self.actor.s3_conn.get_bucket_tagging.return_value = {
            'TagSet': [
                {'Key': 'k1', 'Value': 'v1'},
                {'Key': 'k2', 'Value': 'v2'}
            ]}
        ret = yield self.actor._get_tags()
        self.assertEqual(ret, [
            {'key': 'k1', 'value': 'v1'},
            {'key': 'k2', 'value': 'v2'}
        ])

    @testing.gen_test
    def test_get_tags_no_bucket(self):
        self.actor._bucket_exists = False
        ret = yield self.actor._get_tags()
        self.assertEqual(None, ret)

    @testing.gen_test
    def test_get_tags_not_managed(self):
        self.actor._options['tags'] = None
        ret = yield self.actor._get_tags()
        self.assertEqual(None, ret)

    @testing.gen_test
    def test_get_tags_empty(self):
        self.actor._bucket_exists = True
        self.actor.s3_conn.get_bucket_tagging.side_effect = ClientError(
            {'Error': {'Code': ''}}, 'NoSuchTagSet')
        ret = yield self.actor._get_tags()
        self.assertEqual(ret, [])

    @testing.gen_test
    def test_get_tags_exc(self):
        self.actor._bucket_exists = True
        self.actor.s3_conn.get_bucket_tagging.side_effect = ClientError(
            {'Error': {'Code': ''}}, 'SomeOtherError')
        with self.assertRaises(ClientError):
            yield self.actor._get_tags()

    @testing.gen_test
    def test_compare_tags(self):
        self.actor._bucket_exists = True
        self.actor._options['tags'] = [
            {'key': 'test_key', 'value': 'test_value'}
        ]
        self.actor.s3_conn.get_bucket_tagging.return_value = {
            'TagSet': [
                {'Key': 'test_key', 'Value': 'test_value'},
            ]}
        ret = yield self.actor._compare_tags()
        self.assertTrue(ret)

    @testing.gen_test
    def test_compare_tags_false(self):
        self.actor._bucket_exists = True
        self.actor._options['tags'] = [
            {'key': 'test_key', 'value': 'test_value'}
        ]
        self.actor.s3_conn.get_bucket_tagging.return_value = {
            'TagSet': [
                {'Key': 'k1', 'Value': 'v1'},
                {'Key': 'k2', 'Value': 'v2'}
            ]}
        ret = yield self.actor._compare_tags()
        self.assertFalse(ret)

    @testing.gen_test
    def test_compare_tags_not_managing(self):
        self.actor._bucket_exists = True
        self.actor._options['tags'] = None
        ret = yield self.actor._compare_tags()
        self.assertTrue(ret)

    @testing.gen_test
    def test_set_tags_none(self):
        self.actor._options['tags'] = None
        yield self.actor._set_tags()
        self.assertFalse(self.actor.s3_conn.put_bucket_tagging.called)

    @testing.gen_test
    def test_set_tags(self):
        self.actor._options['tags'] = [
            {'key': 'tag1', 'value': 'v1'}
        ]
        yield self.actor._set_tags()
        self.actor.s3_conn.put_bucket_tagging.assert_has_calls([
            mock.call(
                Bucket='test',
                Tagging={'TagSet': [
                    {'Key': 'tag1', 'Value': 'v1'}
                ]})])
